{-# OPTIONS_GHC -fglasgow-exts #-}

module Polynomial where {

import List;
import ErrorMonad;
import HOL;
import Option;
import Product_Type;
import Show;
import Eithera;

mul_pvarsI ::
  forall a b. (Eq a) => a -> (a -> a -> a) -> a -> [(b, a)] -> [(b, a)];
mul_pvarsI uu uv a [] = [];
mul_pvarsI ze ti a ((x, b) : vas) =
  let {
    p = ti a b;
    res = mul_pvarsI ze ti a vas;
  } in (if p == ze then res else (x, p) : res);

data L_poly a b = LPoly b [(a, b)];

mul_lpolyI ::
  forall a b. (Eq a) => a -> (a -> a -> a) -> a -> L_poly b a -> L_poly b a;
mul_lpolyI ze ti a (LPoly b vas) = LPoly (ti a b) (mul_pvarsI ze ti a vas);

list_sumI :: forall a. a -> (a -> a -> a) -> [a] -> a;
list_sumI ze pl [] = ze;
list_sumI ze pl (x : xs) = pl x (list_sumI ze pl xs);

add_varI ::
  forall a b.
    (Eq a, Eq b) => a -> (a -> a -> a) -> b -> a -> [(b, a)] -> [(b, a)];
add_varI uu uv x a [] = [(x, a)];
add_varI ze pl x a ((y, b) : vas) =
  (if x == y
    then let {
           s = pl a b;
         } in (if s == ze then vas else (x, s) : vas)
    else (y, b) : add_varI ze pl x a vas);

sum_pvarsI ::
  forall a b.
    (Eq a, Eq b) => a -> (a -> a -> a) -> [(b, a)] -> [(b, a)] -> [(b, a)];
sum_pvarsI uu uv [] vbs = vbs;
sum_pvarsI ze pl ((x, a) : vas) vbs =
  (if a == ze then sum_pvarsI ze pl vas vbs
    else sum_pvarsI ze pl vas (add_varI ze pl x a vbs));

sum_lpolyI ::
  forall a b.
    (Eq a,
      Eq b) => a -> (a -> a -> a) -> L_poly b a -> L_poly b a -> L_poly b a;
sum_lpolyI ze pl (LPoly a vas) (LPoly b vbs) =
  LPoly (pl a b) (sum_pvarsI ze pl vas vbs);

l_poly_case :: forall a v t. (a -> [(v, a)] -> t) -> L_poly v a -> t;
l_poly_case f1 (LPoly a list) = f1 a list;

show_pvars :: forall a b. (Eq a, Sshow a, Sshow b) => a -> [(b, a)] -> [[Char]];
show_pvars on [] = [];
show_pvars on ((x, c) : vas) =
  append (if on == c then [] else sshow c) (sshow x) : show_pvars on vas;

show_lpoly ::
  forall a b. (Eq a, Sshow a, Sshow b) => a -> a -> L_poly b a -> [Char];
show_lpoly ze on (LPoly c cs) =
  (case show_pvars on cs of {
    [] -> sshow c;
    s : ss ->
      append
        (append (if c == ze then [] else append (sshow c) [' ', '+', ' ']) s)
        (concata (mapa (append [' ', '+', ' ']) ss));
  });

lookup_rest :: forall a b. (Eq a) => a -> [(a, b)] -> Maybe (b, [(a, b)]);
lookup_rest x [] = Nothing;
lookup_rest x ((y, c) : ycs) =
  (if x == y then Just (c, ycs)
    else (case lookup_rest x ycs of {
           Nothing -> Nothing;
           Just a -> let {
                       (d, yccs) = a;
                     } in Just (d, (y, c) : yccs);
         }));

check_pvarsI ::
  forall a b.
    (Eq b) => a -> (a -> a -> Bool) ->
                     (a -> a -> Bool) -> [(b, a)] -> [(b, a)] -> Eithera b ();
check_pvarsI ze ge co vas [] =
  catcha (checkAllMI (\ va -> checkB (ge (snd va) ze) (fst va)) vas)
    (\ x -> errora (snd x));
check_pvarsI ze ge co vas ((x, b) : vbs) =
  let {
    a = (case lookup_rest x vas of {
          Nothing -> (ze, vas);
          Just a -> let {
                      (aa, ba) = a;
                    } in (aa, ba);
        });
    (a', vas') = a;
  } in bind (checkB (co a' b) x) (\ xa -> check_pvarsI ze ge co vas' vbs);

check_lpoly_sI ::
  forall a b.
    (Eq a, Sshow a, Eq b,
      Sshow b) => Bool ->
                    a -> a -> (a -> a -> Bool) ->
                                (a -> a -> Bool) ->
                                  L_poly b a ->
                                    L_poly b a -> Eithera [[Char]] ();
check_lpoly_sI mode ze on ge gt (LPoly a vas) (LPoly b vbs) =
  catcha
    (bind (bind (checkB mode
                  ['c', 'u', 'r', 'r', 'e', 'n', 't', 'l', 'y', ' ', 'm', 'a',
                    'x', '-', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 's', 'o', 'n',
                    's', ' ', 'a', 'r', 'e', ' ', 'n', 'o', 't', ' ', 's', 'u',
                    'p', 'p', 'o', 'r', 't', 'e', 'd'])
            (\ x ->
              checkB (gt a b)
                ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                  ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c',
                  'o', 'n', 's', 't', 'a', 'n', 't', ' ', 'p', 'a', 'r', 't']))
      (\ x ->
        catcha (check_pvarsI ze ge ge vas vbs)
          (\ xa ->
            errora
              (append
                ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                  ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c',
                  'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', 's', ' ',
                  'o', 'f', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ']
                (sshow xa)))))
    (\ x ->
      errora
        [append
           (append
             (append
               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ',
                 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ']
               (show_lpoly ze on (LPoly a vas)))
             [' ', '>', ' '])
           (show_lpoly ze on (LPoly b vbs)),
          x]);

check_lpoly_nsI ::
  forall a b.
    (Eq a, Sshow a, Eq b,
      Sshow b) => a -> a -> (a -> a -> Bool) ->
                              L_poly b a -> L_poly b a -> Eithera [[Char]] ();
check_lpoly_nsI ze on ge (LPoly a vas) (LPoly b vbs) =
  catcha
    (bind (checkB (ge a b)
            ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ',
              'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c', 'o', 'n',
              's', 't', 'a', 'n', 't', ' ', 'p', 'a', 'r', 't', 's'])
      (\ x ->
        catcha (check_pvarsI ze ge ge vas vbs)
          (\ xa ->
            errora
              (append
                ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n',
                  ' ', 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ', 'c',
                  'o', 'e', 'f', 'f', 'i', 'c', 'i', 'e', 'n', 't', 's', ' ',
                  'o', 'f', ' ', 'v', 'a', 'r', 'i', 'a', 'b', 'l', 'e', ' ']
                (sshow xa)))))
    (\ x ->
      errora
        [append
           (append
             (append
               ['p', 'r', 'o', 'b', 'l', 'e', 'm', ' ', 'w', 'h', 'e', 'n', ' ',
                 'c', 'o', 'm', 'p', 'a', 'r', 'i', 'n', 'g', ' ']
               (show_lpoly ze on (LPoly a vas)))
             [' ', '>', '=', ' '])
           (show_lpoly ze on (LPoly b vbs)),
          x]);

}
