{-# OPTIONS_GHC -fglasgow-exts #-}

module RedPairImpl where {

import qualified Nat;
import qualified List;
import qualified Show;
import qualified PolyOrder;
import qualified ErrorMonad;
import qualified Term;
import qualified Eithera;
import qualified Orderings;
import qualified Integer;

data Arctic = MinInfty | Nat_arc Nat.Nat;

data Redpair a = Nat_carrier [((a, Nat.Nat), (Nat.Nat, [Nat.Nat]))]
  | Int_carrier [((a, Nat.Nat), (Integer, [Integer]))]
  | Arctic_carrier [((a, Nat.Nat), (RedPairImpl.Arctic, [RedPairImpl.Arctic]))];

eq_arctic :: RedPairImpl.Arctic -> RedPairImpl.Arctic -> Bool;
eq_arctic (RedPairImpl.Nat_arc a) RedPairImpl.MinInfty = False;
eq_arctic RedPairImpl.MinInfty (RedPairImpl.Nat_arc a) = False;
eq_arctic (RedPairImpl.Nat_arc nat) (RedPairImpl.Nat_arc nat') =
  Nat.eq_nat nat nat';
eq_arctic RedPairImpl.MinInfty RedPairImpl.MinInfty = True;

instance Eq RedPairImpl.Arctic where {
  a == b = RedPairImpl.eq_arctic a b;
};

int_name :: [Char];
int_name =
  ['n', 'a', 't', 'u', 'r', 'a', 'l', 's', ' ', 'w', 'i', 't', 'h', ' ', 'n',
    'e', 'g', 'a', 't', 'i', 'v', 'e', ' ', 'c', 'o', 'n', 's', 't', 'a', 'n',
    't', 's'];

nat_name :: [Char];
nat_name = ['n', 'a', 't', 'u', 'r', 'a', 'l', 's'];

gt_arctic :: RedPairImpl.Arctic -> RedPairImpl.Arctic -> Bool;
gt_arctic RedPairImpl.MinInfty uu = False;
gt_arctic (RedPairImpl.Nat_arc uv) RedPairImpl.MinInfty = True;
gt_arctic (RedPairImpl.Nat_arc x) (RedPairImpl.Nat_arc y) = Nat.less_nat y x;

geq_arctic :: RedPairImpl.Arctic -> RedPairImpl.Arctic -> Bool;
geq_arctic xa RedPairImpl.MinInfty = True;
geq_arctic RedPairImpl.MinInfty (RedPairImpl.Nat_arc uu) = False;
geq_arctic (RedPairImpl.Nat_arc x) (RedPairImpl.Nat_arc y) =
  Nat.less_eq_nat y x;

arctic_name :: [Char];
arctic_name =
  ['a', 'r', 'c', 't', 'i', 'c', ' ', 's', 'e', 'm', 'i', 'r', 'i', 'n', 'g'];

mult_arctic :: RedPairImpl.Arctic -> RedPairImpl.Arctic -> RedPairImpl.Arctic;
mult_arctic RedPairImpl.MinInfty ya = RedPairImpl.MinInfty;
mult_arctic (RedPairImpl.Nat_arc v) RedPairImpl.MinInfty = RedPairImpl.MinInfty;
mult_arctic (RedPairImpl.Nat_arc x) (RedPairImpl.Nat_arc y) =
  RedPairImpl.Nat_arc (Nat.plus_nat x y);

plus_arctic :: RedPairImpl.Arctic -> RedPairImpl.Arctic -> RedPairImpl.Arctic;
plus_arctic RedPairImpl.MinInfty ya = ya;
plus_arctic (RedPairImpl.Nat_arc v) RedPairImpl.MinInfty =
  RedPairImpl.Nat_arc v;
plus_arctic (RedPairImpl.Nat_arc x) (RedPairImpl.Nat_arc y) =
  RedPairImpl.Nat_arc (Orderings.maxa x y);

sshow_arctic :: RedPairImpl.Arctic -> [Char];
sshow_arctic (RedPairImpl.Nat_arc x) = Show.sshow_nat x;
sshow_arctic RedPairImpl.MinInfty = ['-', 'i', 'n', 'f'];

instance Show.Sshow RedPairImpl.Arctic where {
  sshow = RedPairImpl.sshow_arctic;
};

create_rule_removal_proc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => RedPairImpl.Redpair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             Eithera.Eithera [[Char]] ();
create_rule_removal_proc (RedPairImpl.Int_carrier ia) =
  PolyOrder.create_rule_removal_proc_comp
    ((\ x -> True),
      (0, (1, ((\ a b -> a + b),
                ((\ a b -> a * b),
                  (Orderings.maxa 0,
                    ((\ x y -> y <= x),
                      ((\ x y -> y < x), (True, RedPairImpl.int_name)))))))))
    (\ a -> 0 < a) ia;
create_rule_removal_proc (RedPairImpl.Nat_carrier i) =
  PolyOrder.create_rule_removal_proc_comp
    ((\ x -> True),
      (Nat.Zero_nat,
        (Nat.one_nat,
          (Nat.plus_nat,
            (Nat.times_nat,
              (id, ((\ x y -> Nat.less_eq_nat y x),
                     ((\ x y -> Nat.less_nat y x),
                       (True, RedPairImpl.nat_name)))))))))
    (\ x -> not (x == Nat.Zero_nat)) i;
create_rule_removal_proc (RedPairImpl.Arctic_carrier ib) =
  (\ uu uua ->
    ErrorMonad.errora
      [['a', 'r', 'c', 't', 'i', 'c', ' ', 'n', 'o', 't', ' ', 'm', 'o', 'n',
         'o', 't', 'o', 'n', 'i', 'c', ' ', 'f', 'o', 'r', ' ', 'T', 'R', 'S',
         's']]);

create_mono_redpair_ur_proc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => RedPairImpl.Redpair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             [(Term.Term a b, Term.Term a b)] ->
                               Maybe (a -> [(Term.Term a b, Term.Term a b)]) ->
                                 [(Term.Term a b, Term.Term a b)] ->
                                   [(Term.Term a b, Term.Term a b)] ->
                                     Eithera.Eithera [[Char]] ();
create_mono_redpair_ur_proc (RedPairImpl.Int_carrier ia) =
  PolyOrder.create_mono_redpair_ur_proc_comp
    ((\ x -> True),
      (0, (1, ((\ a b -> a + b),
                ((\ a b -> a * b),
                  (Orderings.maxa 0,
                    ((\ x y -> y <= x),
                      ((\ x y -> y < x), (True, RedPairImpl.int_name)))))))))
    (\ a -> 0 < a) ia;
create_mono_redpair_ur_proc (RedPairImpl.Nat_carrier i) =
  PolyOrder.create_mono_redpair_ur_proc_comp
    ((\ x -> True),
      (Nat.Zero_nat,
        (Nat.one_nat,
          (Nat.plus_nat,
            (Nat.times_nat,
              (id, ((\ x y -> Nat.less_eq_nat y x),
                     ((\ x y -> Nat.less_nat y x),
                       (True, RedPairImpl.nat_name)))))))))
    (\ x -> not (x == Nat.Zero_nat)) i;
create_mono_redpair_ur_proc (RedPairImpl.Arctic_carrier ib) =
  (\ uu uua uub uuc uud uue ->
    ErrorMonad.errora
      [['a', 'r', 'c', 't', 'i', 'c', ' ', 'n', 'o', 't', ' ', 'm', 'o', 'n',
         'o', 't', 'o', 'n', 'i', 'c', ' ', 'f', 'o', 'r', ' ', 'T', 'R', 'S',
         's']]);

create_mono_redpair_proc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => RedPairImpl.Redpair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             [(Term.Term a b, Term.Term a b)] ->
                               [(Term.Term a b, Term.Term a b)] ->
                                 Eithera.Eithera [[Char]] ();
create_mono_redpair_proc (RedPairImpl.Int_carrier ia) =
  PolyOrder.create_mono_redpair_proc_comp
    ((\ x -> True),
      (0, (1, ((\ a b -> a + b),
                ((\ a b -> a * b),
                  (Orderings.maxa 0,
                    ((\ x y -> y <= x),
                      ((\ x y -> y < x), (True, RedPairImpl.int_name)))))))))
    (\ a -> 0 < a) ia;
create_mono_redpair_proc (RedPairImpl.Nat_carrier i) =
  PolyOrder.create_mono_redpair_proc_comp
    ((\ x -> True),
      (Nat.Zero_nat,
        (Nat.one_nat,
          (Nat.plus_nat,
            (Nat.times_nat,
              (id, ((\ x y -> Nat.less_eq_nat y x),
                     ((\ x y -> Nat.less_nat y x),
                       (True, RedPairImpl.nat_name)))))))))
    (\ x -> not (x == Nat.Zero_nat)) i;
create_mono_redpair_proc (RedPairImpl.Arctic_carrier ib) =
  (\ uu uua uub uuc ->
    ErrorMonad.errora
      [['a', 'r', 'c', 't', 'i', 'c', ' ', 'n', 'o', 't', ' ', 'm', 'o', 'n',
         'o', 't', 'o', 'n', 'i', 'c', ' ', 'f', 'o', 'r', ' ', 'T', 'R', 'S',
         's']]);

create_redpair_ur_af_proc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => RedPairImpl.Redpair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             [(Term.Term a b, Term.Term a b)] ->
                               Maybe (a -> [(Term.Term a b, Term.Term a b)]) ->
                                 [(Term.Term a b, Term.Term a b)] ->
                                   Eithera.Eithera [[Char]] ();
create_redpair_ur_af_proc (RedPairImpl.Int_carrier ia) =
  PolyOrder.create_redpair_ur_af_proc_comp
    ((\ x -> True),
      (0, (1, ((\ a b -> a + b),
                ((\ a b -> a * b),
                  (Orderings.maxa 0,
                    ((\ x y -> y <= x),
                      ((\ x y -> y < x), (True, RedPairImpl.int_name)))))))))
    ia;
create_redpair_ur_af_proc (RedPairImpl.Nat_carrier i) =
  PolyOrder.create_redpair_ur_af_proc_comp
    ((\ x -> True),
      (Nat.Zero_nat,
        (Nat.one_nat,
          (Nat.plus_nat,
            (Nat.times_nat,
              (id, ((\ x y -> Nat.less_eq_nat y x),
                     ((\ x y -> Nat.less_nat y x),
                       (True, RedPairImpl.nat_name)))))))))
    i;
create_redpair_ur_af_proc (RedPairImpl.Arctic_carrier ib) =
  PolyOrder.create_redpair_ur_af_proc_comp
    ((\ x -> True),
      (RedPairImpl.MinInfty,
        (RedPairImpl.Nat_arc Nat.Zero_nat,
          (RedPairImpl.plus_arctic,
            (RedPairImpl.mult_arctic,
              (id, (RedPairImpl.geq_arctic,
                     (RedPairImpl.gt_arctic,
                       (False, RedPairImpl.arctic_name)))))))))
    ib;

create_redpair_proc ::
  forall a b.
    (Eq a, Show.Sshow a, Eq b,
      Show.Sshow b) => RedPairImpl.Redpair a ->
                         [(Term.Term a b, Term.Term a b)] ->
                           [(Term.Term a b, Term.Term a b)] ->
                             [(Term.Term a b, Term.Term a b)] ->
                               Eithera.Eithera [[Char]] ();
create_redpair_proc (RedPairImpl.Int_carrier ia) =
  PolyOrder.create_redpair_proc_comp
    ((\ x -> True),
      (0, (1, ((\ a b -> a + b),
                ((\ a b -> a * b),
                  (Orderings.maxa 0,
                    ((\ x y -> y <= x),
                      ((\ x y -> y < x), (True, RedPairImpl.int_name)))))))))
    ia;
create_redpair_proc (RedPairImpl.Nat_carrier i) =
  PolyOrder.create_redpair_proc_comp
    ((\ x -> True),
      (Nat.Zero_nat,
        (Nat.one_nat,
          (Nat.plus_nat,
            (Nat.times_nat,
              (id, ((\ x y -> Nat.less_eq_nat y x),
                     ((\ x y -> Nat.less_nat y x),
                       (True, RedPairImpl.nat_name)))))))))
    i;
create_redpair_proc (RedPairImpl.Arctic_carrier ib) =
  PolyOrder.create_redpair_proc_comp
    ((\ x -> True),
      (RedPairImpl.MinInfty,
        (RedPairImpl.Nat_arc Nat.Zero_nat,
          (RedPairImpl.plus_arctic,
            (RedPairImpl.mult_arctic,
              (id, (RedPairImpl.geq_arctic,
                     (RedPairImpl.gt_arctic,
                       (False, RedPairImpl.arctic_name)))))))))
    ib;

}
