(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Matrix imports PolyOrder
begin




(* matrices *)

types 'a vec = "'a list"
types 'a mat = "'a vec list" (* list of column-vectors *)

definition wf_vecI :: "nat \<Rightarrow> 'x vec \<Rightarrow> bool"
 where "wf_vecI n x = (length x = n)"

definition wf_mat_rcI :: "nat \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> bool" where
 "wf_mat_rcI nr nc m = (length m = nc \<and> forall (wf_vecI nr) m)"

lemma wf_mat_def: "wf_mat_rcI nr nc m = (length m = nc \<and> (\<forall> v \<in> set m. wf_vecI nr v))"
  unfolding wf_mat_rcI_def by auto

definition vec_zeroI :: "'a \<Rightarrow> nat \<Rightarrow> 'a vec" where 
 "vec_zeroI ze n = replicate n ze"

definition mat_zeroI :: "'a \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a mat" where
  "mat_zeroI ze nr nc = replicate nc (vec_zeroI ze nr)"

definition vec_plusI :: "('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a vec" where 
 "vec_plusI pl v w = map (\<lambda> xy. pl (fst xy) (snd xy)) (zip v w)"


definition mat_plusI :: "('a \<Rightarrow> 'a \<Rightarrow> 'a) \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
 where "mat_plusI pl m1 m2 = map (\<lambda> uv. vec_plusI pl (fst uv) (snd uv)) (zip m1 m2)"

definition vec_geI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
  where "vec_geI ge v w \<equiv> forall (\<lambda> (a,b). ge a b) (zip v w)"

definition mat_geI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
  where "mat_geI ge m1 m2 \<equiv> forall (\<lambda> (v,w). vec_geI ge v w) (zip m1 m2)"

definition vec_pre_gtI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
  where "vec_pre_gtI gt v w \<equiv> any (\<lambda> (a,b). gt a b) (zip v w)"

definition mat_pre_gtI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
  where "mat_pre_gtI gt m1 m2 \<equiv> any (\<lambda> (v,w). vec_pre_gtI gt v w) (zip m1 m2)"

definition mat_gtI :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> 'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
  where "mat_gtI ge gt m1 m2 \<equiv> mat_geI ge m1 m2 \<and> mat_pre_gtI gt m1 m2"

definition vec_max0I :: "('a \<Rightarrow> 'a) \<Rightarrow> 'a vec \<Rightarrow> 'a vec"
  where "vec_max0I = map"

definition vec1I :: "'a \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a vec"
  where "vec1I ze on n i \<equiv> replicate i ze @ on # replicate (n - 1 - i) ze"

definition mat1I :: "'a \<Rightarrow> 'a \<Rightarrow> nat \<Rightarrow> 'a mat"
  where "mat1I ze on n \<equiv> map (vec1I ze on n) [0 ..< n]"

definition mat_max0I :: "('a \<Rightarrow> 'a) \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
  where "mat_max0I mz = map (vec_max0I mz)"

locale matrix_order = 
  fixes n :: nat (* dimension *)
  and zero :: "'a :: sshow" ("0")
  and one :: 'a ("1")
  and plus :: "'a \<Rightarrow> 'a \<Rightarrow> 'a" (infix "+" 65)
  and times :: "'a \<Rightarrow> 'a \<Rightarrow> 'a" (infix "*" 70)
  and geq :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<ge>" 21)
  and gt :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix ">" 21)
  and max0 :: "'a \<Rightarrow> 'a"
  and plus_single_mono :: "bool"
  assumes n_pos: "n > 0" (* otherwise there is no 1-matrix where 1-mat > 0-mat *)
  and plus_single_mono (* currently no arctic support, would require other definition of mat_gt! *)
  and plus_comm: "(x :: 'a) +y = y+x"
  and plus_ass[simp]: "(x :: 'a) +(y+z) = (x+y)+z"
  and zero_left[simp]: "0+(x :: 'a) = x"
  and times_ass[simp]: "(x :: 'a) *(y*z) = (x*y)*z"
  and left_mult_zero[simp]: "0* (x :: 'a) = 0"
  and right_mult_zero[simp]: "(x :: 'a) *0 = 0"
  and distrib_left[simp]: "(x :: 'a) *(y+z) = (x*y)+(x*z)"
  and distrib_right[simp]: "((x :: 'a) +y)*z = (x*z)+(y*z)"
  and compat[trans]: "\<lbrakk>s \<ge> (t :: 'a); t > u\<rbrakk> \<Longrightarrow> s > u"
  and compat2[trans]: "\<lbrakk>s > (t :: 'a); t \<ge> u\<rbrakk> \<Longrightarrow> s > u"
  and plus_left_mono: "\<lbrakk>(x :: 'a) \<ge> y\<rbrakk> \<Longrightarrow> x + z \<ge> y + z"
  and times_left_mono: "\<lbrakk>z \<ge> 0; (x :: 'a) \<ge> y\<rbrakk> \<Longrightarrow> x * z \<ge> y * z"
  and times_right_mono: "\<lbrakk>x \<ge> 0; (y :: 'a) \<ge> z\<rbrakk> \<Longrightarrow> x * y \<ge> x * z"
  and geq_refl[simp]: "x \<ge> (x :: 'a)"
  and geq_trans[trans]: "\<lbrakk>x \<ge> y; (y :: 'a) \<ge> z\<rbrakk> \<Longrightarrow> x \<ge> z"
  and SN: "SN {(x,y) . geq y zero \<and> gt x y}"
  and max0_id_pos: "\<lbrakk>geq x  zero\<rbrakk> \<Longrightarrow> max0 x = x"
  and max0_pos: "geq (max0 x) zero"
  and max0_mono: "\<lbrakk>x \<ge> y\<rbrakk> \<Longrightarrow> geq (max0 x) (max0 y)"
  and max0_x: "geq (max0 x) x"
  and left_one[simp]: "1 * (x :: 'a) = x"
  and right_one[simp]: "(x :: 'a) * 1 = x"
  and one_geq_zero[simp]: "one \<ge> zero"
  and one_gt_zero[simp]: "one > zero"
  and plus_gt_left_mono: "\<lbrakk>(x :: 'a) > y; z \<ge> u; plus_single_mono\<rbrakk> \<Longrightarrow> x + z > y + u"
  and plus_gt_both_mono: "\<lbrakk>(x :: 'a) > y; z > u; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x + z > y + u" 
begin
  
abbreviation wf_vec :: "'a vec \<Rightarrow> bool"
where "wf_vec \<equiv> wf_vecI n"

abbreviation wf_mat :: "'a mat \<Rightarrow> bool"
where "wf_mat \<equiv> wf_mat_rcI n n"

abbreviation vec0 :: "'a vec"
where "vec0 \<equiv> vec_zeroI zero n"

abbreviation mat0 :: "'a mat"
where "mat0 \<equiv> mat_zeroI zero n n"

abbreviation mat1 :: "'a mat"
where "mat1 \<equiv> mat1I zero one n"

abbreviation vec_max0 :: "'a vec \<Rightarrow> 'a vec"
where "vec_max0 \<equiv> vec_max0I max0"

abbreviation mat_max0 :: "'a mat \<Rightarrow> 'a mat"
where "mat_max0 \<equiv> mat_max0I max0"

abbreviation vec_plus :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a vec"
where "vec_plus \<equiv> vec_plusI plus"

abbreviation mat_plus :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat"
where "mat_plus \<equiv> mat_plusI plus"

abbreviation vec_ge :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
where "vec_ge \<equiv> vec_geI geq"

abbreviation vec_pre_gt :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
where "vec_pre_gt \<equiv> vec_pre_gtI gt"

abbreviation vec_gt :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> bool"
where "vec_gt u v \<equiv> vec_ge u v \<and> vec_pre_gt u v"

abbreviation mat_ge :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_ge \<equiv> mat_geI geq"

abbreviation mat_pre_gt :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_pre_gt \<equiv> mat_pre_gtI gt"

abbreviation mat_gt :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> bool"
where "mat_gt \<equiv> mat_gtI geq gt"

lemma plus_right_mono: "\<lbrakk>(y :: 'a) \<ge> z\<rbrakk> \<Longrightarrow> x + y \<ge> x + z"
  by (simp add: plus_comm[where x = x], rule plus_left_mono, auto)

lemma wf_vec_max0: assumes "wf_vec v" shows "wf_vec (vec_max0 v)"
using assms unfolding wf_vecI_def vec_max0I_def by auto

lemma wf_mat_max0: assumes "wf_mat_rcI n nc m" shows "wf_mat_rcI n nc (mat_max0 m)"
using assms unfolding wf_mat_rcI_def mat_max0I_def
  by (auto simp: wf_vec_max0)

lemma vec_max0_x: "vec_ge (vec_max0 v) v"
unfolding vec_geI_def vec_max0I_def
by (auto simp: set_zip max0_x)

lemma mat_max0_x: "mat_ge (mat_max0 m) m"
unfolding mat_geI_def mat_max0I_def
by (auto simp: set_zip vec_max0_x)

lemma vec_max0_pos: assumes "wf_vec v"
  shows "vec_ge (vec_max0 v) vec0"
using assms unfolding wf_vecI_def vec_zeroI_def vec_max0I_def vec_geI_def
by (auto simp: set_zip max0_pos)

lemma mat_max0_pos: assumes "wf_mat_rcI n nc v"
  shows "mat_ge (mat_max0 v) (mat_zeroI zero n nc)"
using assms unfolding wf_mat_rcI_def mat_zeroI_def mat_max0I_def mat_geI_def
by (auto simp: set_zip vec_max0_pos)

lemma vec_max0_id_pos: assumes "vec_ge v vec0" and "wf_vec v"  
  shows "vec_max0 v = v"
using assms unfolding wf_vecI_def vec_zeroI_def vec_max0I_def vec_geI_def 
by (induct v arbitrary: n, auto simp: max0_id_pos)

lemma mat_max0_id_pos: assumes "mat_ge m (mat_zeroI zero n nc)" and "wf_mat_rcI n nc m"  
  shows "mat_max0 m = m"
using assms unfolding wf_mat_rcI_def mat_zeroI_def mat_max0I_def mat_geI_def 
by (induct m arbitrary: nc, auto simp: vec_max0_id_pos)

lemma vec_max0_mono: assumes "vec_ge v w" and "wf_vec v" and "wf_vec w"
  shows "vec_ge (vec_max0 v) (vec_max0 w)"
using assms unfolding wf_vecI_def vec_max0I_def vec_geI_def
proof (simp, induct n arbitrary: v w, simp)
  case (Suc nn)
  show ?case
  proof (cases v, simp)
    case (Cons a v1)
    show ?thesis
    proof (cases w, simp)
      case (Cons b w1)
      from prems have len: "(a \<ge> b) \<and> length v1 = nn \<and> length w1 = nn \<and> (\<forall> a b . (a,b) \<in> set (zip v1 w1) \<longrightarrow> (a \<ge> b))" by auto
      show ?thesis
	by (simp add: `v = a # v1` `w = b # w1`, rule conjI, rule max0_mono, simp add: len, rule Suc, auto simp: len)
    qed
  qed
qed

lemma mat_max0_mono: assumes "mat_ge v w" and "wf_mat_rcI n nc v" and "wf_mat_rcI n nc w"
  shows "mat_ge (mat_max0 v) (mat_max0 w)"
using assms unfolding wf_mat_rcI_def mat_max0I_def mat_geI_def
proof (simp, induct nc arbitrary: v w, simp)
  case (Suc nn)
  show ?case
  proof (cases v, simp)
    case (Cons a v1)
    show ?thesis
    proof (cases w, simp)
      case (Cons b w1)
      from prems have len: "(vec_ge a b) \<and> wf_vec a \<and> wf_vec b \<and> length v1 = nn \<and> length w1 = nn \<and> (\<forall> a \<in> set v1. wf_vec a) \<and> (\<forall> b \<in> set w1. wf_vec b) \<and> (\<forall> a b . (a,b) \<in> set (zip v1 w1) \<longrightarrow> (vec_ge a b))" by auto
      show ?thesis
	by (simp add: `v = a # v1` `w = b # w1`, rule conjI, rule vec_max0_mono, (simp add: len)+, rule Suc, auto simp: len)
    qed
  qed
qed


lemma wf_vec0[simp] : "wf_vec vec0"
  by (simp add: wf_vecI_def vec_zeroI_def)

lemma wf_mat0[simp] : "wf_mat mat0"
  by (simp add: wf_mat_rcI_def mat_zeroI_def, rule replicate_prop, simp)

lemma vec_plus_comm : "vec_plus x y = vec_plus y x"
unfolding vec_plusI_def
proof (induct x arbitrary: y, simp)
  case (Cons a x)
  thus ?case 
    by (cases y, simp, auto simp: plus_comm) 
qed

lemma mat_plus_comm : "mat_plus m1 m2 = mat_plus m2 m1"
unfolding mat_plusI_def
proof (induct m1 arbitrary: m2, simp)
  case (Cons v m1)
  thus ?case
  proof (cases m2, simp)
    case (Cons w m2a)
    hence "mat_plus (v # m1) m2 = vec_plus v w # mat_plus m1 m2a" by (auto simp: mat_plusI_def)
    also have "\<dots> = vec_plus w v # mat_plus m1 m2a" using vec_plus_comm by auto
    finally show ?thesis using prems by (auto simp: mat_plusI_def)
  qed
qed

lemma vec_plus_assoc[simp]: assumes "wf_vec u" and "wf_vec v" and "wf_vec w"
 shows "vec_plus u (vec_plus v w) = vec_plus (vec_plus u v) w"
using assms
unfolding wf_vecI_def 
proof (induct n arbitrary: u v w, simp add: vec_plusI_def)
 case (Suc nn)
 thus ?case 
 proof (cases u, simp, cases v, simp, cases w, simp)
   fix x uu y vv z ww
   assume "u = x # uu" and "v = y # vv" and "w = z # ww"
   with Suc show "vec_plus u (vec_plus v w) = vec_plus (vec_plus u v) w" by (simp add: vec_plusI_def)
 qed
qed

lemma mat_plus_assoc[simp]: assumes "wf_mat_rcI n nc m1" and "wf_mat_rcI n nc m2" and "wf_mat_rcI n nc m3"
  shows "mat_plus m1 (mat_plus m2 m3) = mat_plus (mat_plus m1 m2) m3"
using assms
proof (induct nc arbitrary: m1 m2 m3, simp add: mat_plusI_def wf_mat_rcI_def)
  case (Suc nn)
  thus ?case
  proof (cases m1, simp add: wf_mat_rcI_def, cases m2, simp add: wf_mat_rcI_def, cases m3, simp add: wf_mat_rcI_def)
    fix v1 mm1 v2 mm2 v3 mm3
    assume 
      "wf_mat_rcI n (Suc nn) m1" and m1: "m1 = v1 # mm1" and
      "wf_mat_rcI n (Suc nn) m2" and m2: "m2 = v2 # mm2" and
      "wf_mat_rcI n (Suc nn) m3" and m3: "m3 = v3 # mm3"
    hence v1: "wf_vec v1" and v2: "wf_vec v2" and v3: "wf_vec v3" by (auto simp: wf_mat_rcI_def)
    hence id: "vec_plus v1 (vec_plus v2 v3) = vec_plus (vec_plus v1 v2) v3" by (rule vec_plus_assoc)
    from m1 m2 m3 have "mat_plus m1 (mat_plus m2 m3) = vec_plus v1 (vec_plus v2 v3) # mat_plus mm1 (mat_plus mm2 mm3)" by (auto simp: mat_plusI_def)
    also have "\<dots> = vec_plus (vec_plus v1 v2) v3 # mat_plus mm1 (mat_plus mm2 mm3)" by (simp only: id)
    finally show "mat_plus m1 (mat_plus m2 m3) = mat_plus (mat_plus m1 m2) m3" using prems unfolding wf_mat_rcI_def mat_plusI_def by auto 
  qed
qed

lemma vec0_plus[simp]: assumes "wf_vec u" shows "vec_plus vec0 u = u"
using assms
unfolding wf_vecI_def vec_plusI_def vec_zeroI_def
proof (induct n arbitrary: u, simp)
 case (Suc nn) thus ?case by (cases u, auto)
qed

lemma mat0_plus[simp]: assumes "wf_mat_rcI n nc m" shows "mat_plus (mat_zeroI zero n nc) m = m"
using assms 
unfolding wf_mat_rcI_def 
proof (induct nc arbitrary: m, simp add: mat_plusI_def mat_zeroI_def)
  case (Suc nn) 
  thus ?case proof (cases m, simp)
    case (Cons v mm)
    with Suc have wf: "wf_vec v" by auto
    from Cons Suc have "mat_plus (mat_zeroI 0 n (Suc nn)) m = vec_plus vec0 v # mat_plus (mat_zeroI 0 n nn) mm" by (auto simp: mat_plusI_def mat_zeroI_def)
    also have "\<dots> = vec_plus vec0 v # mm" using Suc Cons by auto
    also have "\<dots> = v # mm" by (simp only: vec0_plus wf)
    finally show ?thesis using Cons by auto
  qed
qed

lemma wf_vec_plus : "\<lbrakk>wf_vec u; wf_vec v\<rbrakk> \<Longrightarrow> wf_vec (vec_plus u v)"
using assms 
unfolding vec_plusI_def wf_vecI_def
by (induct n arbitrary: u v, auto)

lemma wf_mat_plus : assumes "wf_mat_rcI n nc m1" and "wf_mat_rcI n nc m2" shows "wf_mat_rcI n nc (mat_plus m1 m2)"
using assms
unfolding wf_mat_def mat_plusI_def
proof (simp, induct nc arbitrary: m1 m2, simp)
  case (Suc nn)
  show ?case 
  proof (cases m1)
    case Nil with Suc show ?thesis by auto
  next
    case (Cons v1 mm1)
    with Suc have l1: "length mm1 = nn" by auto
    show ?thesis
    proof (cases m2)
      case Nil with Suc show ?thesis by auto
    next
      case (Cons v2 mm2)
      with Suc have l2: "length mm2 = nn" by auto
      show ?thesis by  (simp add: prems, rule conjI, rule wf_vec_plus, (simp add: prems)+, rule Suc, auto simp: prems l1 l2)
    qed
  qed
qed


lemma vec_plus_left_mono: "\<lbrakk>vec_ge v1 v2; wf_vec v1; wf_vec v2; wf_vec v3\<rbrakk> \<Longrightarrow> vec_ge (vec_plus v1 v3) (vec_plus v2 v3)"
unfolding vec_geI_def vec_plusI_def wf_vecI_def
proof (simp, induct n arbitrary: v1 v2 v3, simp)
  case (Suc nn)
  show ?case 
  proof (cases v1, simp, cases v2, simp, cases v3, simp)
    fix a1 w1 a2 w2 a3 w3
    assume v1: "v1 = a1 # w1" and v2: "v2 = a2 # w2" and v3: "v3 = a3 # w3"    
    with Suc have geq: "a1 \<ge> a2" by auto
    from v1 v2 v3 Suc have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn" by auto
    show "\<forall> a \<in> set (zip (map (\<lambda> xy. fst xy + snd xy) (zip v1 v3)) 
                         (map (\<lambda> xy. fst xy + snd xy) (zip v2 v3))). split op \<ge> a"
      by (simp add: v1 v2 v3, rule conjI, rule plus_left_mono, rule geq, rule Suc, auto simp: Suc v1 v2 v3 len)
  qed
qed

lemma mat_plus_left_mono: assumes ge: "mat_ge m1 m2" and wf1: "wf_mat_rcI n nc m1" and wf2: "wf_mat_rcI n nc m2" and wf3: "wf_mat_rcI n nc m3" 
  shows "mat_ge (mat_plus m1 m3) (mat_plus m2 m3)"
proof -
  from ge have "\<forall> (v,w) \<in> set (zip m1 m2). vec_ge v w" unfolding mat_geI_def by auto
  with wf1 wf2 wf3 show "mat_ge (mat_plus m1 m3) (mat_plus m2 m3)" unfolding mat_geI_def mat_plusI_def wf_mat_def
  proof (simp, induct nc arbitrary: m1 m2 m3, simp)
    case (Suc nn)
    show ?case 
    proof (cases m1, simp, cases m2, simp, cases m3, simp)
      fix v1 mm1 v2 mm2 v3 mm3
      assume v1: "m1 = v1 # mm1" and v2: "m2 = v2 # mm2" and v3: "m3 = v3 # mm3"    
      with Suc have geq: "vec_ge v1 v2" by auto
      from v1 v2 v3 Suc have len: "length mm1 = nn \<and> length mm2 = nn \<and> length mm3 = nn" by auto
      show "\<forall> a \<in> set (zip (map (\<lambda> xy. vec_plus (fst xy) (snd xy)) (zip m1 m3)) 
                      (map (\<lambda> xy. vec_plus (fst xy) (snd xy)) (zip m2 m3))). split vec_ge a"
	by (simp add: v1 v2 v3, rule conjI, rule vec_plus_left_mono, rule geq, simp add: Suc v1, simp add: Suc v2, simp add: Suc v3, rule Suc,
	  auto simp: len v1 v2 v3 Suc)
    qed
  qed
qed
      
lemma vec_ge_refl: "vec_ge v v"
  unfolding vec_geI_def 
  by (induct v, auto)

lemma mat_ge_refl: "mat_ge m m"
  unfolding mat_geI_def
  by (induct m, auto simp: vec_ge_refl)

lemma vec_ge_trans: assumes "vec_ge v1 v2" and "vec_ge v2 v3" and "wf_vec v1" and "wf_vec v2" and "wf_vec v3" 
  shows "vec_ge v1 v3"
unfolding vec_geI_def
using assms unfolding wf_vecI_def
proof (simp, induct n arbitrary: v1 v2 v3, simp)
  case (Suc nn)
  thus ?case 
  proof (cases v1)
    case Nil with Suc show ?thesis by auto
  next
    case (Cons a1 w1)
    show ?thesis
    proof (cases v2)
      case Nil with Suc show ?thesis by auto
    next
      case (Cons a2 w2)
      show ?thesis
      proof (cases v3)
	case Nil with Suc show ?thesis by auto
      next
	case (Cons a3 w3)
	from `vec_ge v1 v2` `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `vec_ge v2 v3` 
	have a12: "a1 \<ge> a2" and a23: "a2 \<ge> a3" and w12: "vec_ge w1 w2" and w23: "vec_ge w2 w3" unfolding vec_geI_def by auto
	from prems have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn" by auto 
	show ?thesis 
	  by (simp add: `v1 = a1 # w1` Cons, rule conjI, rule geq_trans, rule a12, rule a23, rule Suc, rule w12, rule w23, auto simp: len w12 w23)
      qed
    qed
  qed
qed


lemma mat_ge_trans: assumes "mat_ge v1 v2" and "mat_ge v2 v3" and "wf_mat_rcI n nc v1" and "wf_mat_rcI n nc v2" and "wf_mat_rcI n nc v3" 
  shows "mat_ge v1 v3"
unfolding mat_geI_def 
using assms 
proof (simp, induct nc arbitrary: v1 v2 v3, simp add: wf_mat_rcI_def)
  case (Suc nn)
  thus ?case 
  proof (cases v1)
    case Nil with Suc show ?thesis by auto
  next
    case (Cons a1 w1)
    show ?thesis
    proof (cases v2)
      case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
    next
      case (Cons a2 w2)
      show ?thesis
      proof (cases v3)
	case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
      next
	case (Cons a3 w3)
	from `mat_ge v1 v2` `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `mat_ge v2 v3` 
	have a12: "vec_ge a1 a2" and a23: "vec_ge a2 a3" and w12: "mat_ge w1 w2" and w23: "mat_ge w2 w3" unfolding mat_geI_def by auto
	from `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `wf_mat_rcI n (Suc nn) v1` `wf_mat_rcI n (Suc nn) v2` `wf_mat_rcI n (Suc nn) v3` 
	have wf: "wf_mat_rcI n nn w1 \<and> wf_mat_rcI n  nn w2 \<and> wf_mat_rcI n nn w3 \<and> wf_vec a1 \<and> wf_vec a2 \<and> wf_vec a3" unfolding wf_mat_def by auto
	from prems have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn" unfolding wf_mat_def by auto 
	show ?thesis 
	  by (simp add: `v1 = a1 # w1` Cons, rule conjI, rule vec_ge_trans, rule a12, rule a23, (simp add: wf)+, rule Suc, rule w12, rule w23, auto simp: wf, rule w12, rule w23)
      qed
    qed
  qed
qed

lemma vec_pre_compat: assumes "vec_ge v1 v2" and "vec_pre_gt v2 v3" and "wf_vec v1" and "wf_vec v2" and "wf_vec v3" 
  shows "vec_pre_gt v1 v3"
unfolding vec_pre_gtI_def
using assms unfolding wf_vecI_def 
proof (simp add: any_mem_conv mem_iff, induct n arbitrary: v1 v2 v3, simp add: vec_pre_gtI_def)
  case (Suc nn)
  thus ?case 
  proof (cases v1)
    case Nil with Suc show ?thesis by auto
  next
    case (Cons a1 w1)
    show ?thesis
    proof (cases v2)
      case Nil with Suc show ?thesis by auto
    next
      case (Cons a2 w2)
      show ?thesis
      proof (cases v3)
	case Nil with Suc show ?thesis by auto
      next
	case (Cons a3 w3)
	from `vec_ge v1 v2` `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `vec_pre_gt v2 v3` 
	have a12: "a1 \<ge> a2" and w12: "vec_ge w1 w2" and w23: "(a2 > a3) \<or> vec_pre_gt w2 w3" unfolding vec_geI_def vec_pre_gtI_def by (cases "a2 > a3", auto)
	from prems have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn" by auto 
	from w23 show ?thesis
	proof
	  assume "a2 > a3"
	  with a12 have "a1 > a3" by (rule compat)
	  from prems have "(a1,a3) \<in> set (zip v1 v3)" by auto
	  with `a1 > a3` show ?thesis by blast
	next
	  assume gt: "vec_pre_gt w2 w3"
	  have "\<exists> a b. (a,b) \<in> set (zip w1 w3) \<and> (a > b)"
	    by (rule Suc, rule `vec_ge w1 w2`, auto simp: len w12 gt)
	  thus ?thesis by (auto simp: `v1 = a1 # w1` `v3 = a3 # w3`)
	qed
      qed
    qed
  qed
qed

lemma vec_pre_compat2: assumes "vec_pre_gt v1 v2" and "vec_ge v2 v3" and "wf_vec v1" and "wf_vec v2" and "wf_vec v3" 
  shows "vec_pre_gt v1 v3"
unfolding vec_pre_gtI_def
using assms unfolding wf_vecI_def 
proof (simp add: any_mem_conv mem_iff, induct n arbitrary: v1 v2 v3, simp add: vec_pre_gtI_def)
  case (Suc nn)
  thus ?case 
  proof (cases v1)
    case Nil with Suc show ?thesis by auto
  next
    case (Cons a1 w1)
    show ?thesis
    proof (cases v2)
      case Nil with Suc show ?thesis by auto
    next
      case (Cons a2 w2)
      show ?thesis
      proof (cases v3)
	case Nil with Suc show ?thesis by auto
      next
	case (Cons a3 w3)
	from `vec_pre_gt v1 v2` `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `vec_ge v2 v3` 
	have a12: "a2 \<ge> a3" and w12: "vec_ge w2 w3" and w23: "(a1 > a2) \<or> vec_pre_gt w1 w2" unfolding vec_geI_def vec_pre_gtI_def by (cases "a1 > a2", auto)
	from prems have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn" by auto 
	from w23 show ?thesis
	proof
	  assume "a1 > a2"
	  from this and a12 have "a1 > a3" by (rule compat2)
	  from prems have "(a1,a3) \<in> set (zip v1 v3)" by auto
	  with `a1 > a3` show ?thesis by blast
	next
	  assume gt: "vec_pre_gt w1 w2"
	  have "\<exists> a b. (a,b) \<in> set (zip w1 w3) \<and> (a > b)"
	    by (rule Suc, rule gt, rule w12, auto simp: len w12 gt)
	  thus ?thesis by (auto simp: `v1 = a1 # w1` `v3 = a3 # w3`)
	qed
      qed
    qed
  qed
qed

lemma mat_pre_compat: assumes "mat_ge v1 v2" and "mat_pre_gt v2 v3" and "wf_mat_rcI n nc v1" and "wf_mat_rcI n nc v2" and "wf_mat_rcI n nc v3" 
  shows "mat_pre_gt v1 v3"
unfolding mat_pre_gtI_def
using assms 
proof (simp add: any_mem_conv mem_iff, induct nc arbitrary: v1 v2 v3, simp add: mat_pre_gtI_def wf_mat_rcI_def)
  case (Suc nn)
  thus ?case 
  proof (cases v1)
    case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
  next
    case (Cons a1 w1)
    show ?thesis
    proof (cases v2)
      case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
    next
      case (Cons a2 w2)
      show ?thesis
      proof (cases v3)
	case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
      next
	case (Cons a3 w3)
	from `mat_ge v1 v2` `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `mat_pre_gt v2 v3` `wf_mat_rcI n (Suc nn) v1` `wf_mat_rcI n (Suc nn) v2` `wf_mat_rcI n (Suc nn) v3`
	have a12: "vec_ge a1 a2" and w12: "mat_ge w1 w2" and w23: "vec_pre_gt a2 a3 \<or> mat_pre_gt w2 w3" 
	  and wf: "wf_mat_rcI n nn w1 \<and> wf_mat_rcI n nn w2 \<and> wf_mat_rcI n nn w3 \<and> wf_vec a1 \<and> wf_vec a2 \<and> wf_vec a3" unfolding mat_geI_def mat_pre_gtI_def wf_mat_rcI_def by (cases "vec_pre_gt a2 a3", auto)
	from w23 show ?thesis
	proof
	  assume gt: "vec_pre_gt a2 a3"
	  have "vec_pre_gt a1 a3" by (rule vec_pre_compat, rule a12, auto simp: gt wf)
	  from prems have "(a1,a3) \<in> set (zip v1 v3)" by auto
	  with `vec_pre_gt a1 a3` show ?thesis by blast
	next
	  assume gt: "mat_pre_gt w2 w3"
	  have "\<exists> a b. (a,b) \<in> set (zip w1 w3) \<and> (vec_pre_gt a b)"
	    by (rule Suc, rule `mat_ge w1 w2`, auto simp: wf w12 gt)
	  thus ?thesis by (auto simp: `v1 = a1 # w1` `v3 = a3 # w3`)
	qed
      qed
    qed
  qed
qed

lemma mat_pre_compat2: assumes "mat_pre_gt v1 v2" and "mat_ge v2 v3" and "wf_mat_rcI n nc v1" and "wf_mat_rcI n nc v2" and "wf_mat_rcI n nc v3" 
  shows "mat_pre_gt v1 v3"
unfolding mat_pre_gtI_def
using assms 
proof (simp add: any_mem_conv mem_iff, induct nc arbitrary: v1 v2 v3, simp add: mat_pre_gtI_def wf_mat_rcI_def)
  case (Suc nn)
  thus ?case 
  proof (cases v1)
    case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
  next
    case (Cons a1 w1)
    show ?thesis
    proof (cases v2)
      case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
    next
      case (Cons a2 w2)
      show ?thesis
      proof (cases v3)
	case Nil with Suc show ?thesis by (auto simp: wf_mat_rcI_def)
      next
	case (Cons a3 w3)
	from `mat_ge v2 v3` `v1 = a1 # w1` `v2 = a2 # w2` `v3 = a3 # w3` `mat_pre_gt v1 v2` `wf_mat_rcI n (Suc nn) v1` `wf_mat_rcI n (Suc nn) v2` `wf_mat_rcI n (Suc nn) v3`
	have a12: "vec_ge a2 a3" and w12: "mat_ge w2 w3" and w23: "vec_pre_gt a1 a2 \<or> mat_pre_gt w1 w2" 
	  and wf: "wf_mat_rcI n nn w1 \<and> wf_mat_rcI n nn w2 \<and> wf_mat_rcI n nn w3 \<and> wf_vec a1 \<and> wf_vec a2 \<and> wf_vec a3" unfolding mat_geI_def mat_pre_gtI_def wf_mat_rcI_def by (cases "vec_pre_gt a1 a2", auto)
	from w23 show ?thesis
	proof
	  assume gt: "vec_pre_gt a1 a2"
	  have "vec_pre_gt a1 a3" by (rule vec_pre_compat2, rule gt, rule a12, auto simp: gt wf)
	  from prems have "(a1,a3) \<in> set (zip v1 v3)" by auto
	  with `vec_pre_gt a1 a3` show ?thesis by blast
	next
	  assume gt: "mat_pre_gt w1 w2"
	  have "\<exists> a b. (a,b) \<in> set (zip w1 w3) \<and> (vec_pre_gt a b)"
	    by (rule Suc, rule gt, rule `mat_ge w2 w3`, auto simp: wf w12 gt)
	  thus ?thesis by (auto simp: `v1 = a1 # w1` `v3 = a3 # w3`)
	qed
      qed
    qed
  qed
qed


lemma mat_gt_compat: assumes ge: "mat_ge v1 v2" and gt: "mat_gt v2 v3" and wf1: "wf_mat_rcI n nc v1" and wf2: "wf_mat_rcI n nc v2" and wf3: "wf_mat_rcI n nc v3" 
  shows "mat_gt v1 v3"
proof -
  from gt have ge1: "mat_ge v2 v3" and gt1: "mat_pre_gt v2 v3" unfolding mat_gtI_def by auto
  show ?thesis unfolding mat_gtI_def
    by (rule conjI, rule mat_ge_trans, rule ge, rule ge1, rule wf1, auto simp: wf2 wf3, rule mat_pre_compat, rule ge, rule gt1, rule wf1, auto simp: wf2 wf3)
qed  

lemma mat_gt_compat2: assumes gt: "mat_gt v1 v2" and ge: "mat_ge v2 v3" and wf1: "wf_mat_rcI n nc v1" and wf2: "wf_mat_rcI n nc v2" and wf3: "wf_mat_rcI n nc v3" 
  shows "mat_gt v1 v3"
proof -
  from gt have ge1: "mat_ge v1 v2" and gt1: "mat_pre_gt v1 v2" unfolding mat_gtI_def by auto
  show ?thesis unfolding mat_gtI_def
    by (rule conjI, rule mat_ge_trans, rule ge1, rule ge, rule wf1, auto simp: wf2 wf3, rule mat_pre_compat2, rule gt1, rule ge, rule wf1, auto simp: wf2 wf3)
qed


lemma vec_plus_pre_gt_left_mono: assumes gt: "vec_pre_gt v1 v2" and ge: "vec_ge v3 v4" and wf1: "wf_vec v1" and wf2: "wf_vec v2" and wf3: "wf_vec v3" and wf4: "wf_vec v4"
  and single: "plus_single_mono"
  shows "vec_pre_gt (vec_plus v1 v3) (vec_plus v2 v4)"
proof - 
  from gt ge wf1 wf2 wf3 wf4
  show "vec_pre_gt (vec_plus v1 v3) (vec_plus v2 v4)" unfolding vec_pre_gtI_def wf_vecI_def vec_plusI_def vec_geI_def
  proof (induct n arbitrary: v1 v2 v3 v4, simp)
    case (Suc nn)
    hence "length v1 = Suc nn \<and> length v2 = Suc nn \<and> length v3 = Suc nn \<and> length v4 = Suc nn" by auto
    thus ?case
    proof (cases v1, simp, cases v2, simp, cases v3, simp, cases v4, simp)
      fix a1 w1 a2 w2 a3 w3 a4 w4
      assume v1: "v1 = a1 # w1" and v2: "v2 = a2 # w2" and v3: "v3 = a3 # w3" and v4: "v4 = a4 # w4"
      with Suc have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn \<and> length w4 = nn" by auto
      let ?pl = "\<lambda> u1 u2. map (\<lambda> xy. plus (fst xy) (snd xy)) (zip u1 u2)"
      show "any (\<lambda>(a, b). a > b) (zip (?pl v1 v3) (?pl v2 v4))"
      proof (cases "a1 > a2")
	case True
	from Suc v1 v2 v3 v4 have ge: "a3 \<ge> a4" by auto
	have "a1 + a3 > a2 + a4" by (rule plus_gt_left_mono, auto simp: True ge single)
	with v1 v2 v3 v4 show ?thesis by auto
      next
	case False
	with Suc v1 v2 v3 v4 have tp: "any (\<lambda>(a,b). a > b) (zip w1 w2)" and ta: "forall (\<lambda>(a,b). a \<ge> b) (zip w3 w4)" by auto
	have "any (\<lambda>(a, b). a > b) (zip (?pl w1 w3) (?pl w2 w4))" 
	  by (rule Suc, rule tp, rule ta, auto simp: len)
	with v1 v2 v3 v4 False show ?thesis by auto
      qed
    qed
  qed
qed	
    


lemma mat_plus_pre_gt_left_mono: assumes gt: "mat_pre_gt v1 v2" and ge: "mat_ge v3 v4" and wf1: "wf_mat_rcI n nc v1" and wf2: "wf_mat_rcI n nc v2" and wf3: "wf_mat_rcI n nc v3" and wf4: "wf_mat_rcI n nc v4"
  and single: "plus_single_mono"
  shows "mat_pre_gt (mat_plus v1 v3) (mat_plus v2 v4)"
proof -
  from gt ge wf1 wf2 wf3 wf4
  show "mat_pre_gt (mat_plus v1 v3) (mat_plus v2 v4)" unfolding mat_pre_gtI_def wf_mat_rcI_def mat_plusI_def mat_geI_def
  proof (induct nc arbitrary: v1 v2 v3 v4, simp)
    case (Suc nn)
    hence "length v1 = Suc nn \<and> length v2 = Suc nn \<and> length v3 = Suc nn \<and> length v4 = Suc nn" by auto
    thus ?case
    proof (cases v1, simp, cases v2, simp, cases v3, simp, cases v4, simp)
      fix a1 w1 a2 w2 a3 w3 a4 w4
      assume v1: "v1 = a1 # w1" and v2: "v2 = a2 # w2" and v3: "v3 = a3 # w3" and v4: "v4 = a4 # w4"
      with Suc have len: "length w1 = nn \<and> length w2 = nn \<and> length w3 = nn \<and> length w4 = nn" by auto
      from Suc v1 v2 v3 v4 have wf: "wf_vec a1 \<and> wf_vec a2 \<and> wf_vec a3 \<and> wf_vec a4 \<and> 
	forall wf_vec w1 \<and> forall wf_vec w2 \<and> forall wf_vec w3 \<and> forall wf_vec w4 \<and>
	vec_ge a3 a4 \<and> forall (\<lambda> (a,b). vec_ge a b) (zip w3 w4)" by auto
      let ?pl = "\<lambda> u1 u2. map (\<lambda> xy. vec_plus (fst xy) (snd xy)) (zip u1 u2)"
      show "any (\<lambda>(a, b). vec_pre_gt a b) (zip (?pl v1 v3) (?pl v2 v4))"
      proof (cases "vec_pre_gt a1 a2")
	case True
	from Suc v1 v2 v3 v4 have ge: "vec_ge a3 a4" by auto
	have "vec_pre_gt (vec_plus a1 a3) (vec_plus a2 a4)" by (rule vec_plus_pre_gt_left_mono, auto simp: True single wf)
	with v1 v2 v3 v4 show ?thesis by auto
      next
	case False
	with Suc v1 v2 v3 v4 have tp: "any (\<lambda>(a,b). vec_pre_gt a b) (zip w1 w2)" by auto
	have "any (\<lambda>(a, b). vec_pre_gt a b) (zip (?pl w1 w3) (?pl w2 w4))" 
	  by (rule Suc, rule tp, auto simp only: len wf)
	with v1 v2 v3 v4 False show ?thesis by auto
      qed
    qed
  qed
qed	

lemma mat_plus_gt_left_mono: assumes gt: "mat_gt v1 v2" and ge: "mat_ge v3 v4" and wf1: "wf_mat_rcI n nc v1" and wf2: "wf_mat_rcI n nc v2" and wf3: "wf_mat_rcI n nc v3" and wf4: "wf_mat_rcI n nc v4"
  and single: "plus_single_mono"
  shows "mat_gt (mat_plus v1 v3) (mat_plus v2 v4)"
proof -
  from gt have ge1: "mat_ge v1 v2" and gt1: "mat_pre_gt v1 v2" unfolding mat_gtI_def by auto
  show ?thesis unfolding mat_gtI_def 
    by (rule conjI, rule mat_ge_trans[where ?v2.0 = "mat_plus v2 v3" and nc = nc], rule mat_plus_left_mono[where nc = nc], auto simp: ge1 ge wf1 wf2 wf3 wf4 wf_mat_plus, simp add: mat_plus_comm[where ?m1.0 = v2], rule mat_plus_left_mono[where nc = nc], rule ge, auto simp: wf1 wf2 wf3 wf4,
    rule mat_plus_pre_gt_left_mono[where nc = nc], rule gt1, rule ge, auto simp: wf1 wf2 wf3 wf4 single)
qed


abbreviation vec_ns :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> bool" ("(_ \<succeq>v _)" [51,51] 50) 
 where "v \<succeq>v w \<equiv> (wf_vec v \<and> wf_vec w \<and> vec_ge v w)"

abbreviation vec_s :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> bool" ("(_ \<succ>v _)" [51,51] 50)
 where "v \<succ>v w \<equiv> (wf_vec v \<and> wf_vec w \<and> vec_ge w vec0 \<and> vec_gt v w)"

lemma vec_sn : "SN {(v,w) . v \<succ>v w}"
unfolding SN_ars_def SN_elt_def 
proof clarify
 fix e f 
 assume ass: "\<forall> i. (f i, f (Suc i)) \<in> {(v,w). v \<succ>v w}"
 hence ns_vec: "\<And> i. (f i \<succeq>v f (Suc i))" by auto
 hence len_n: "\<And> i. length (f i) = n" by (auto simp: wf_vecI_def)
 with ns_vec have ns: "\<And> i j. j < n \<Longrightarrow> f i ! j \<ge> f (Suc i) ! j" by (auto simp: vec_geI_def set_zip)
 have vec0: "\<And> v. wf_vec v \<Longrightarrow> vec_ge v vec0 = (\<forall> i < n. v ! i \<ge> 0)" 
 proof -
   fix v
   assume "wf_vec v" hence n: "n = length v" unfolding wf_vecI_def by auto
   show "vec_ge v vec0 = (\<forall> i < n. v ! i \<ge> 0)"
     unfolding vec_geI_def vec_zeroI_def 
     by (simp only: n, simp add: all_set_conv_all_nth)
 qed   
   
 from ass len_n have "\<And> i. \<exists> j < n. ((wf_vec (f (Suc i)) \<and> vec_ge (f (Suc i)) vec0)) \<and> (f i ! j > f (Suc i) ! j)" unfolding vec_pre_gtI_def by (auto simp: any_mem_conv mem_iff set_zip wf_vecI_def)
 hence s: "\<And> i. \<exists> j < n. ((f (Suc i)) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j)" using vec0 by blast

 show False
 proof (cases "\<exists> j < n. \<forall> m . \<exists> i . i \<ge> m \<and> (f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j)")
   case False
   with False have "\<forall> j < n. \<exists> m . \<forall> i . i \<ge> m \<longrightarrow> \<not> ((f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j))" by blast
   hence "\<exists> m. \<forall> j < n. \<forall> i. i \<ge> m \<longrightarrow> \<not> ((f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j))"
   proof (induct n, simp)
     case (Suc nn)
     from Suc obtain m1 where res1: "\<forall> j < nn. \<forall> i \<ge> m1 . \<not> ((f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j))" by auto
     from Suc obtain m2 where res2: "\<forall> i \<ge> m2. \<not> ((f (Suc i) ! nn \<ge> 0) \<and> (f i ! nn > f (Suc i) ! nn))" by auto
     let ?m = "max m1 m2"
     have "\<forall> j < Suc nn. \<forall> i \<ge> ?m. \<not> ((f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j))" 
     proof (rule allI)
	fix j
	show "j < Suc nn \<longrightarrow> (\<forall> i \<ge> ?m . \<not> ((f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j)))"
	  using res1 res2 by (cases "j = nn", auto)
     qed
     thus ?case ..
   qed
   with s show False by blast      
 next
   case True
   from this obtain j where jlen: "j < n" and jinf: "\<forall> m. \<exists> i \<ge> m. (f (Suc i) ! j \<ge> 0) \<and> (f i ! j > f (Suc i) ! j)" by auto
   
   obtain g where gdef: "g = (\<lambda> i. f i ! j)" by auto
   let ?S = "{(x,y). (y \<ge> 0) \<and> ((x :: 'a) > y)}"
   let ?NS = "{(x,y). (x :: 'a) \<ge> y}"
   from SN have SNstart: "SN ?S (g 0)" unfolding SN_ars_def by simp
   from ns jlen gdef have steps: "\<forall> i. (g i, g (Suc i)) \<in> ?NS \<union> ?S" by simp
   from compat have "?S O ?NS \<subseteq> ?S" by auto
   from steps this SNstart have "\<exists> m. \<forall> i \<ge> m. (g i, g (Suc i)) \<in> ?NS - ?S" by (rule non_strict_ending)
   with gdef jinf show False by blast
 qed
qed

abbreviation mat_ns :: "'a mat \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> bool" ("(_ \<succeq>_ _)" [51,51,51] 50) 
 where "v \<succeq>nc w \<equiv> (wf_mat_rcI n nc v \<and> wf_mat_rcI n nc w \<and> mat_ge v w)"

abbreviation mat_s :: "'a mat \<Rightarrow> nat \<Rightarrow> 'a mat \<Rightarrow> bool" ("(_ \<succ>_ _)" [51,51,51] 50)
 where "v \<succ>nc w \<equiv> (wf_mat_rcI n nc v \<and> wf_mat_rcI n nc w \<and> mat_ge w (mat_zeroI 0 n nc) \<and> mat_gt v w)"

lemma mat_sn : "SN {(v,w) . v \<succ>nc w}"
unfolding SN_ars_def SN_elt_def 
proof clarify
 fix e f 
 assume ass: "\<forall> i. (f i, f (Suc i)) \<in> {(v,w). v \<succ>nc w}"
 hence ns_vec: "\<And> i. (f i \<succeq>nc f (Suc i))" unfolding mat_gtI_def by auto
 hence len_n: "\<And> i. length (f i) = nc" by (auto simp: wf_mat_rcI_def)
 have vec0: "\<And> v. wf_mat_rcI n nc v \<Longrightarrow> mat_ge v (mat_zeroI 0 n nc) = (\<forall> i < nc. vec_ge (v ! i) vec0)" 
 proof -
   fix v :: "'a mat"
   assume "wf_mat_rcI n nc v" hence n: "nc = length v" unfolding wf_mat_rcI_def by auto
   show "mat_ge v (mat_zeroI 0 n nc) = (\<forall> i < nc. vec_ge (v ! i) vec0)"
     unfolding mat_geI_def mat_zeroI_def 
     by (simp only: n, simp add: all_set_conv_all_nth)
 qed   
 have ns: "\<And> i j. j < nc \<Longrightarrow> wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f i ! j) (f (Suc i) ! j)" 
 proof -
   fix i j
   assume j: "j < nc"
   have len: "length (f i) = nc" using len_n by auto
   have len2: "length (f (Suc i)) = nc" using len_n by auto
   have vec: "(f i \<succeq>nc f (Suc i))" using ns_vec by auto
   with j have "vec_ge (f i ! j) (f (Suc i) ! j)" unfolding mat_geI_def 
     by (simp add: all_set_conv_all_nth len len2)
   with len vec j show "wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f i ! j) (f (Suc i) ! j)" 
     unfolding wf_mat_rcI_def by auto
 qed
   
 have s: "\<And> i. \<exists> j < nc. (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))"
 proof -
   fix i
   have len: "length (f i) = nc" using len_n by auto
   have len2: "length (f (Suc i)) = nc" using len_n by auto
   from ass have stri: "f i \<succ>nc  f (Suc i)" by auto
   hence "mat_gt (f i) (f (Suc i))" by auto
   from this obtain j where j: "j < nc" and gt: "vec_gt (f i ! j) (f (Suc i) ! j)" unfolding mat_gtI_def mat_geI_def mat_pre_gtI_def
     by (auto simp: any_mem_conv mem_iff set_zip all_set_conv_all_nth len len2)
   from stri j have ze: "vec_ge (f (Suc i) ! j) vec0" by (auto simp: vec0)
   with gt stri j have "wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j)" 
     unfolding wf_mat_rcI_def by auto
   with j show "\<exists> j < nc. (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))" by blast
 qed
   
 show False
 proof (cases "\<exists> j < nc. \<forall> m . \<exists> i . i \<ge> m \<and> wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j)")
   case False
   with False have "\<forall> j < nc. \<exists> m . \<forall> i . i \<ge> m \<longrightarrow> \<not> (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))" by blast
   hence "\<exists> m. \<forall> j < nc. \<forall> i. i \<ge> m \<longrightarrow> \<not> (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))"
   proof (induct nc, simp)
     case (Suc nn)
     from Suc obtain m1 where res1: "\<forall> j < nn. \<forall> i \<ge> m1 . \<not> (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))" by auto
     from Suc obtain m2 where res2: "\<forall> i \<ge> m2. \<not> (wf_vec (f i ! nn) \<and> wf_vec (f (Suc i) ! nn) \<and> vec_ge (f (Suc i) ! nn) vec0 \<and> vec_gt (f i ! nn) (f (Suc i) ! nn))" by auto
     let ?m = "max m1 m2"
     have "\<forall> j < Suc nn. \<forall> i \<ge> ?m. \<not> (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))"
     proof (rule allI)
	fix j
	show "j < Suc nn \<longrightarrow> (\<forall> i \<ge> ?m . \<not> (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j)))"
	  using res1 res2 by (cases "j = nn", auto)
     qed
     thus ?case by blast
   qed
   with s show False by blast      
 next
   case True
   from this obtain j where jlen: "j < nc" and jinf: "\<forall> m. \<exists> i \<ge> m. (wf_vec (f i ! j) \<and> wf_vec (f (Suc i) ! j) \<and> vec_ge (f (Suc i) ! j) vec0 \<and> vec_gt (f i ! j) (f (Suc i) ! j))" by auto
   
   obtain g where gdef: "g = (\<lambda> i. f i ! j)" by auto
   let ?S = "{(x,y). wf_vec x \<and> wf_vec y \<and> vec_ge y vec0 \<and> vec_gt x y}"
   let ?NS = "{(x,y). wf_vec x \<and> wf_vec y \<and> vec_ge x y}"
   from vec_sn have SNstart: "SN ?S (g 0)" unfolding SN_ars_def by simp
   from ns jlen gdef have steps: "\<forall> i. (g i, g (Suc i)) \<in> ?NS \<union> ?S" by simp
   have "?S O ?NS \<subseteq> ?S" 
   proof (clarify)
     fix a b x y z
     assume x: "wf_vec x" and y: "wf_vec y" and z: "wf_vec z" and xy: "vec_ge x y" and yzg: "vec_pre_gt y z" and yz: "vec_ge y z"
     show "vec_ge x z \<and> vec_pre_gt x z"
       by (rule conjI, rule vec_ge_trans, rule xy, rule yz, auto simp: x y z, rule vec_pre_compat, rule xy, rule yzg, auto simp: x y z)
   qed     
   from steps this SNstart have "\<exists> m. \<forall> i \<ge> m. (g i, g (Suc i)) \<in> ?NS - ?S" by (rule non_strict_ending)
   with gdef jinf show False by blast
 qed
qed


lemma wf_vec1: assumes "i < n" shows "wf_vec (vec1I zero one n i)"
unfolding wf_vecI_def vec1I_def using assms by auto

lemma wf_mat1: "wf_mat mat1"
unfolding wf_mat_rcI_def mat1I_def using wf_vec1 by auto

lemma vec1_ge_vec0: assumes "i < n" shows "vec_ge (vec1I zero one n i) vec0"
unfolding vec_geI_def vec_zeroI_def vec1I_def using assms
proof (auto simp: set_zip)
  fix j
  assume i: "i < n" and j: "j < n"
  show "(replicate i 0 @ 1 # replicate (n - Suc i) 0) ! j \<ge> (0 :: 'a)" (is "?repl j \<ge> (0 :: 'a)")
  proof (cases "j < i", simp add: nth_append)
    case False
    show ?thesis
    proof (cases "j - i", simp add: nth_append)
      case (Suc k)
      with False
      have repl: "?repl j = (replicate (n - Suc i) 0) ! k" by (simp add: nth_append)
      from Suc False j have "k < (n - Suc i)" by arith
      with repl show ?thesis by auto
    qed
  qed
qed

lemma vec1_pre_gt_vec0: assumes "i < n" shows "vec_pre_gt (vec1I zero one n i) vec0"
unfolding vec_pre_gtI_def vec_zeroI_def vec1I_def using assms
proof (auto simp: set_zip any_mem_conv mem_iff)
  let ?one = "\<lambda> j. (replicate i 0 @ 1 # replicate (n - Suc i) (0 :: 'a)) ! j"
  let ?zero = "\<lambda> i. (replicate n (0 :: 'a)) ! i"
  have gt: "?one i > ?zero i" using assms by (simp add: nth_append)
  with assms show "\<exists> a b. (\<exists> j. a = ?one j \<and> b = ?zero j \<and> j < n) \<and> (a > b)" by blast 
qed

lemma mat1_ge_mat0: shows "mat_ge mat1 mat0"
unfolding mat_geI_def mat_zeroI_def mat1I_def
by (auto simp: set_zip vec1_ge_vec0)

lemma mat1_pre_gt_mat0: shows "mat_pre_gt mat1 mat0"
proof -
  show ?thesis unfolding mat_pre_gtI_def mat_zeroI_def mat1I_def
  proof (auto simp: any_mem_conv mem_iff set_zip)
    let ?one = "\<lambda> i. map (vec1I (0 :: 'a) 1 n) [0..<n] ! i"
    let ?zero = "\<lambda> i. (replicate n vec0) ! i"
    have gt: "vec_pre_gt (?one 0) (?zero 0)" using n_pos by (simp add: vec1_pre_gt_vec0)
    with n_pos show "\<exists> a b. (\<exists> j. a = ?one j \<and> b = ?zero j \<and> j < n) \<and> (vec_pre_gt a b)" by blast 
  qed
qed

lemma mat1_gt_mat0: shows "mat_gt mat1 mat0"
unfolding mat_gtI_def using mat1_ge_mat0 mat1_pre_gt_mat0 by auto  



lemma "poly_order_carrier wf_mat mat0 mat_plus mat_times mat_ge mat_gt plus_single_mono mat1 mat_max0"
proof (rule poly_order_carrier.intro, rule poly_carrier.intro, rule wf_mat0, rule wf_mat_plus, simp+, rule mat_plus_comm, rule mat_plus_assoc, simp+)
oops




(*
locale poly_order_carrier = poly_carrier + 
  fixes plus_single_mono :: "bool"

  missing lemmas

  and wf_times[simp]: "\<lbrakk>wf_elem x; wf_elem y\<rbrakk> \<Longrightarrow> wf_elem (x*y)"
  and times_ass[simp]: "\<lbrakk>wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x*(y*z) = (x*y)*z"
  and left_mult_zero[simp]: "\<lbrakk>wf_elem x\<rbrakk> \<Longrightarrow> 0*x = 0"
  and right_mult_zero[simp]: "\<lbrakk>wf_elem x\<rbrakk> \<Longrightarrow> x*0 = 0"
  and distrib_left[simp]: "\<lbrakk>wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x*(y+z) = (x*y)+(x*z)"
  and distrib_right[simp]: "\<lbrakk>wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> (x+y)*z = (x*z)+(y*z)"
  and times_left_mono: "\<lbrakk>z \<ge> 0; x \<ge> y; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x * z \<ge> y * z"
  and times_right_mono: "\<lbrakk>x \<ge> 0; y \<ge> z; wf_elem x; wf_elem y; wf_elem z\<rbrakk> \<Longrightarrow> x * y \<ge> x * z"
  and plus_gt_both_mono: "\<lbrakk>x > y; z > u; wf_elem x; wf_elem y; wf_elem z; wf_elem u; \<not> plus_single_mono\<rbrakk> \<Longrightarrow> x + z > y + u"
    and left_one[simp]: "wf_elem x \<Longrightarrow> 1 * x = x"
    and right_one[simp]: "wf_elem x \<Longrightarrow> x * 1 = x"
*)

(*
old stuff 


types 'a mat = "'a vec list"

locale vecops = vecorder + 
 fixes plus :: "'a \<Rightarrow> 'a \<Rightarrow> 'a"
 and   zero :: "'a"
 and   mul :: "'a \<Rightarrow> 'a \<Rightarrow> 'a"
 assumes plus_mono : "\<lbrakk>x1 \<ge> x2; y1 \<ge> y2\<rbrakk> \<Longrightarrow> plus x1 y1 \<ge> plus x2 y2"
 and   mul_mono : "\<lbrakk>x1 \<ge> x2; y1 \<ge> y2\<rbrakk> \<Longrightarrow> mul x1 y1 \<ge> mul x2 y2"
 and   geq_refl : "(x :: 'a) \<ge> x"
begin

fun scalar_prod :: "'a vec \<Rightarrow> 'a vec \<Rightarrow> 'a" where
 "scalar_prod v w = foldr (\<lambda> xy s. plus (mul (fst xy) (snd xy)) s) (zip v w) zero"

lemma scalar_prod_mono : assumes "v1 \<succeq>n v2" and "w1 \<succeq>n w2" shows "scalar_prod v1 w1 \<ge> scalar_prod v2 w2"
using assms
proof (induct n arbitrary: v1 v2 w1 w2, simp add: geq_refl)
 case (Suc nn)
 thus ?case proof (cases v1, simp, cases v2, simp, cases w1, simp, cases w2, simp)
   fix x1 vv1 x2 vv2 y1 ww1 y2 ww2
   assume eq1: "v1 = x1 # vv1" and eq2: "v2 = x2 # vv2" 
   with Suc have x: "x1 \<ge> x2" and vv: "vv1 \<succeq>nn vv2" by auto
   assume eq3: "w1 = y1 # ww1" and eq4: "w2 = y2 # ww2" 
   with Suc have y: "y1 \<ge> y2" and ww: "ww1 \<succeq>nn ww2" by auto
   from vv ww Suc have sc: "scalar_prod vv1 ww1 \<ge> scalar_prod vv2 ww2" by auto
   from x y mul_mono have mul: "mul x1 y1 \<ge> mul x2 y2" by blast
   from eq1 eq2 eq3 eq4 have "scalar_prod v1 w1 = plus (mul x1 y1) (scalar_prod vv1 ww1) \<and> 
     scalar_prod v2 w2 = plus (mul x2 y2) (scalar_prod vv2 ww2)" by auto
   with sc mul plus_mono[where ?x1.0 = "mul x1 y1"] show "scalar_prod v1 w1 \<ge> scalar_prod v2 w2" by auto
 qed
qed

lemma scalar_prod_mono_l : assumes "wf_vec n w" and "v1 \<succeq>n v2" shows "scalar_prod v1 w \<ge> scalar_prod v2 w"
proof (rule scalar_prod_mono)
 from assms vec_geq_refl show "w \<succeq>n w" by auto
next
 from assms show "v1 \<succeq>n v2" by simp
qed

fun mat_vec_mul :: "'a mat \<Rightarrow> 'a vec \<Rightarrow> 'a vec"
 where "mat_vec_mul m v = map (scalar_prod v) m" 

lemma mat_vec_mul_mono : assumes "wf_mat_rc nr nc m" and "v \<succeq>nc w" shows "mat_vec_mul m v \<succeq>nr (mat_vec_mul m w)" 
 using assms by (auto simp: scalar_prod_mono_l simp del: scalar_prod.simps)

end


locale matops = vecops +
 fixes one :: "'a" 
 and   max0 :: "'a \<Rightarrow> 'a" 
 assumes plus_comm: "plus (x :: 'a) y = plus y x"
 and     mul_comm: "mul (x :: 'a) y = mul y x"
 and     plus_assoc[simp]: "plus (plus x y) z = plus x (plus y z)"
 and     mul_distrib_r[simp]: "mul x (plus y z) = plus (mul x y) (mul x z)"
 and     mul_distrib_l[simp]: "mul (plus x y) z = plus (mul x z) (mul y z)"
 and     mul_zero_r[simp]: "mul x zero = zero"
 and     mul_zero_l[simp]: "mul zero x = zero"
 and     plus_zero_r[simp]: "plus x zero = x"
 and     plus_zero_l[simp]: "plus zero x = x"

context matops
begin

lemma scalar_prod_comm:
assumes "wf_vec n v" and "wf_vec n w"
shows "scalar_prod v w = scalar_prod w v"
proof-
from assms have len:"length v = length w" by auto
have "scalar_prod v w = foldr (\<lambda> xy s. plus (mul (fst xy) (snd xy)) s) (zip v w) zero" by auto
also have "\<dots> =  foldr (\<lambda> xy s. plus (mul (snd xy) (fst xy)) s) (zip v w) zero" using mul_comm by auto
also have "\<dots> = foldr (\<lambda> xy s. plus (mul (fst xy) (snd xy)) s) (zip w v) zero" using len by (induct v w rule:list_induct2, simp_all)
finally show ?thesis by auto
qed



lemma vec_plus_zero_r[simp]: assumes "wf_vec n u" shows "vec_plus u (vec_zero n) = u"
using assms
proof (induct n arbitrary: u, simp)
 case (Suc nn) thus ?case by (cases u, auto)
qed


lemma scalar_prod_zero_l[simp]: assumes "wf_vec n u" shows "scalar_prod (vec_zero n) u = zero"
proof (induct n arbitrary: u, simp)
 case (Suc nn)
 thus ?case by (cases u, auto)
qed

lemma scalar_prod_zero_r[simp]: assumes "wf_vec n u" shows "scalar_prod u (vec_zero n) = zero"
proof (induct n arbitrary: u, simp)
 case (Suc nn)
 thus ?case by (cases u, auto)
qed

lemma vec_mul_zero[simp]: assumes "wf_mat_rc nr nc m" shows "mat_vec_mul m (vec_zero nc) = vec_zero nr"
using assms 
proof (induct nr arbitrary: m, simp)
 case (Suc nrr) thus ?case 
 proof (cases m, simp)
   case (Cons v mm)
   with Suc have one: "scalar_prod (vec_zero nc) v = zero" by (auto simp del: vec_zero.simps scalar_prod.simps)
   from Cons Suc have "mat_vec_mul mm (vec_zero nc) = vec_zero nrr" by force
   with one Suc Cons show ?thesis by auto
 qed
qed

lemma scalar_plus_distrib_l[simp]: assumes "wf_vec n u" and "wf_vec n v" and "wf_vec n w" 
 shows "scalar_prod (vec_plus u v) w = plus (scalar_prod u w) (scalar_prod v w)"
using assms
proof (induct n arbitrary: u v w, simp)
 case (Suc nn)
 thus ?case proof (cases u, simp, cases v, simp, cases w, simp)
   fix x uu y vv z ww
   assume eq1: "u = x # uu" and eq2: "v = y # vv" and eq3: "w = z # ww"
   from eq1 eq2 eq3 have "scalar_prod (vec_plus u v) w = plus ( plus (mul x z) (mul y z)) (scalar_prod (vec_plus uu vv) ww)" by simp
   also have "\<dots> = plus ( plus (mul x z) (mul y z)) (plus (scalar_prod uu ww) (scalar_prod vv ww))" using Suc eq1 eq2 eq3 by simp
   also have "\<dots> = plus (plus (mul x z) (scalar_prod uu ww))  (plus (mul y z) (scalar_prod vv ww))" using plus_comm plus_assoc by arith
   also have "\<dots> = plus (scalar_prod u w) (scalar_prod v w)" using Suc eq1 eq2 eq3 by simp
   finally show ?thesis .
 qed
qed

lemma scalar_plus_distrib_r[simp]: assumes "wf_vec n u" and "wf_vec n v" and "wf_vec n w" 
 shows "scalar_prod u (vec_plus v w) = plus (scalar_prod u v) (scalar_prod u w)"
using assms
proof (induct n arbitrary: u v w, simp)
 case (Suc nn)
 thus ?case proof (cases u, simp, cases v, simp, cases w, simp)
   fix x uu y vv z ww
   assume eq1: "u = x # uu" and eq2: "v = y # vv" and eq3: "w = z # ww"
   from eq1 eq2 eq3 have "scalar_prod u (vec_plus v w) = plus ( plus (mul x y) (mul x z)) (scalar_prod uu (vec_plus vv ww))" by simp
   also have "\<dots> = plus ( plus (mul x y) (mul x z)) (plus (scalar_prod uu vv) (scalar_prod uu ww))" using Suc eq1 eq2 eq3 by simp
   also have "\<dots> = plus (plus (mul x y) (scalar_prod uu vv))  (plus (mul x z) (scalar_prod uu ww))" using plus_comm plus_assoc by arith
   also have "\<dots> = plus (scalar_prod u v) (scalar_prod u w)" using Suc eq1 eq2 eq3 by simp
   finally show ?thesis .
 qed
qed

declare scalar_prod.simps[simp del]


lemma mat_vec_mul_distrib[simp]: assumes "wf_mat_rc nr nc m" and "wf_vec nc u" and "wf_vec nc v" 
 shows "mat_vec_mul m (vec_plus u v) = vec_plus (mat_vec_mul m u) (mat_vec_mul m v)"
using assms
proof (induct nr arbitrary: m, simp)
 case (Suc nrr) thus ?case proof (cases m, simp)
   case (Cons w mm)
   from Cons Suc have one: "mat_vec_mul mm (vec_plus u v) = vec_plus (mat_vec_mul mm u) (mat_vec_mul mm v)" by force
   from Cons Suc have two: "scalar_prod (vec_plus u v) w = plus (scalar_prod u w) (scalar_prod v w)" by (auto simp del: vec_plus.simps)
   from Cons Suc one two show "mat_vec_mul m (vec_plus u v) = vec_plus (mat_vec_mul m u) (mat_vec_mul m v)" by auto
 qed
qed

lemma mat_vec_mul_zero[simp]: assumes "wf_vec nc u" shows "mat_vec_mul (mat_zero nr nc) u = vec_zero nr"
using assms
proof (induct nr, simp)
 case (Suc nrr)
 hence one: "mat_vec_mul (mat_zero nrr nc) u = vec_zero nrr" by auto
 from Suc have "scalar_prod u (vec_zero nc) = zero" by (simp only: scalar_prod_zero_r)
 with one Suc show ?case by simp
qed


lemma mat_vec_mul_plus_distrib : assumes wf1: "wf_mat_rc nr nc m1" and wf2: "wf_mat_rc nr nc m2" and wf3: "wf_vec nc w"
 shows "mat_vec_mul (mat_plus m1 m2) w = vec_plus (mat_vec_mul m1 w) (mat_vec_mul m2 w)"
using wf1 wf2
proof (induct nr arbitrary: m1 m2, simp)
 case (Suc nrr) thus ?case proof (cases m1, simp, cases m2, simp)
   fix u m11 v m22
   assume eq1: "m1 = u # m11" and eq2: "m2 = v # m22" 
   from eq1 eq2 Suc have "wf_mat_rc nrr nc m11" and "wf_mat_rc nrr nc m22" by auto
   with Suc have one: "mat_vec_mul (mat_plus m11 m22) w = vec_plus (mat_vec_mul m11 w) (mat_vec_mul m22 w)" by auto
   from eq1 eq2 Suc have "wf_vec nc u" and "wf_vec nc v" by auto
   with wf3 Suc have two: "scalar_prod w (vec_plus u v) = plus (scalar_prod w u) (scalar_prod w v)" by (auto simp del: vec_plus.simps)
   from eq1 eq2 Suc one two show "mat_vec_mul (mat_plus m1 m2) w = vec_plus (mat_vec_mul m1 w) (mat_vec_mul m2 w)" by auto
 qed
qed
(* alte transpose-funktion:

function transpose :: "'a mat \<Rightarrow> 'a mat"
 where "transpose [] = []"
    |  "transpose ([] # m) = []"
    |  "transpose ((x # u) # m) = (let (c,mm) = foldr (\<lambda> v (d,mmm). (hd v # d, tl v # mmm)) m ([],[]) in ((x # c) # transpose mm))"
by (pat_completeness, auto)

termination transpose oops
*)


fun transpose :: "'a mat \<Rightarrow> 'a mat"
 where "transpose [] = []"
    |  "transpose ([] # m) = [[]]"
    |  "transpose ((x # u) # m) = (
          let mm = transpose m in
          case mm of [] \<Rightarrow> (x # []) # (map (\<lambda> y. y # []) u)
          | (Cons v mm')\<Rightarrow> (x # v) # (map (\<lambda> (y, w). y # w) (zip u mm')))"



lemma wf_mat_transpose : 
"wf_mat_rc (Suc nr) (Suc nc) m \<Longrightarrow> wf_mat_rc (Suc nc) (Suc nr) (transpose m)" 
proof (induct m arbitrary: nr, simp)
case (Cons v m)
 assume IH:"\<And> r. wf_mat_rc (Suc r) (Suc nc) m \<Longrightarrow> wf_mat_rc (Suc nc) (Suc r) (transpose m)" and 
        wf:"wf_mat_rc (Suc nr) (Suc nc) (v # m)" 
from wf have "length v = (Suc nc)" by auto
hence "\<exists>x v'. v = x # v' \<and> length v' = nc" by (auto simp:length_Suc_conv)
then obtain x v' where xu:"v = x # v'" and lenv':"length v' = nc" by auto
show ?case
 proof (cases nr)
  case 0
   with wf have "transpose m = []" by auto
   with xu have transp:"transpose (v # m) = (x # []) # (map (\<lambda> y. y # []) v')" by auto
   with lenv' have tlen:"length (transpose (v # m)) = Suc nc" by auto
   have "\<forall> i < (Suc nc). length ((transpose (v # m)) ! i) = (Suc 0)"
    proof
     fix i from transp tlen show "i < Suc nc \<longrightarrow> length (transpose (v # m) ! i) = Suc 0" by (induct i, auto)
    qed
   with tlen 0 show "wf_mat_rc (Suc nc) (Suc nr) (transpose (v # m))" by auto
  next
  case (Suc r)
   from wf have "wf_mat_rc nr (Suc nc) m" by auto
   with IH Suc have wft:"wf_mat_rc (Suc nc) (Suc r) (transpose m)" by auto
   hence "length (transpose m) = Suc nc" by auto
   hence "\<exists>u m'. (transpose m) = u # m' \<and> length m' = nc" by (auto simp:length_Suc_conv)
   then obtain u m' where tm:"(transpose m) = u # m'" and lenm':"length m' = nc" by auto
   with xu have transp:"transpose (v # m) = (x # u) # (map (\<lambda> (y, w). y # w) (zip v' m'))" by auto
   with lenm' lenv' have lent:"length (transpose (v # m)) = (Suc nc)" by auto
   let ?mm="map (\<lambda> (y, w). y # w) (zip v' m')"
   from wft Suc tm have alli:"\<forall> i < (Suc nc). length ((u # m') ! i) = nr" by auto
   with lenm' lenv' have  alli2:"\<forall> i < nc. length (?mm ! i) = Suc nr"  by auto
   from alli have "length (x # u) = Suc nr" by auto
   have "\<forall> i < (Suc nc). length (((x # u) # ?mm) ! i) = (Suc nr)"
    proof
     fix i
     from `length (x # u) = Suc nr` alli2 show " i < Suc nc \<longrightarrow> length (((x # u) # ?mm) ! i) = Suc nr" by (cases i, simp_all)
    qed
   with transp lent show "wf_mat_rc (Suc nc) (Suc nr) (transpose (v # m))" by auto
  qed
 qed

fun rvec_mat_mul :: "'a vec \<Rightarrow> 'a mat \<Rightarrow> 'a vec"
 where "rvec_mat_mul v m = map (scalar_prod v) (transpose m)" 

lemma scalar_prod_mono_r : assumes "wf_vec n w" and "v1 \<succeq>n v2" shows "scalar_prod w v1 \<ge> scalar_prod w v2"
proof (rule scalar_prod_mono)
 from assms vec_geq_refl show "w \<succeq>n w" by auto
next
 from assms show "v1 \<succeq>n v2" by simp
qed


fun mat_mul :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat" 
where "mat_mul m1 m2 = map (\<lambda> v. rvec_mat_mul v m2) m1"

(* alte multiplikation:
fun mat_mul :: "'a mat \<Rightarrow> 'a mat \<Rightarrow> 'a mat" 
where "mat_mul m1 m2 = (let m3 = transpose m2 in map (mat_vec_mul m3) m1)"
*)

(* funktionen um skalarprodukt als summe über einträge aufzufassen *)
fun asum :: "(nat \<Rightarrow> 'a) \<Rightarrow> nat \<Rightarrow> 'a" where
"asum f n = foldl (\<lambda> s i. plus s (f i)) zero [0 ..<n]" 

lemma distr_asum: "mul c (asum f n) = asum (\<lambda> i. mul c (f i)) n" by (induct n, auto)

lemma asum_assoc: shows "foldl (\<lambda> s i. plus s (f i)) (plus x y) zs = plus x (foldl (\<lambda> s i. plus s (f i)) y zs)"
by (induct zs arbitrary: y, simp_all add:plus_assoc)

lemma asum_assoc_r: shows "foldr (\<lambda> i s. plus (f i) s) zs (plus x y) = plus (foldr (\<lambda> i s. plus (f i) s) zs x) y"
by (induct zs arbitrary: y, simp_all add:plus_assoc)

lemma asum_absorb0:
shows "plus x (foldl (\<lambda> s i. plus s (f i)) zero zs) = foldl (\<lambda> s i. plus s (f i)) x zs"
by (induct zs) (simp_all add:asum_assoc)

lemma asum_absorb0_r:
shows "plus (foldr (\<lambda> i s. plus (f i) s) zs zero) x = foldr (\<lambda> i s. plus (f i) s) zs x"
by (induct zs) (simp_all add:asum_assoc_r)

lemma asum_foldr_lemma:
"foldl (\<lambda> s i. plus s (f i)) a xs = plus a (foldr (\<lambda> i s. plus (f i) s) xs zero)"
by (induct xs arbitrary: a, auto simp:plus_assoc plus_comm)

lemma asum_foldr:"asum f n = foldr (\<lambda> i s. plus (f i) s) [0 ..<n] zero" using asum_foldr_lemma[where ?xs="[0..<n]"] by auto

lemma foldr_absorb: shows "foldr (\<lambda> i s. plus (f i) s) zs (plus x y) = plus (foldr (\<lambda> i s. plus (f i) s) zs x) y"
by (induct zs arbitrary: y, simp_all add:plus_assoc)


lemma list_app_last:
assumes "Suc n = length xs"
shows "\<exists> y ys. xs = ys @ [y] \<and> length ys = n" 
proof-
from assms have "Suc n = length (List.rev xs)" by auto
hence "\<exists> z zs. (List.rev xs) = z # zs \<and> length zs = n" using Suc_length_conv[where ?xs = "List.rev xs"] by auto
then obtain z zs where "(List.rev xs) = z # zs" and "length zs = n" by auto
hence "xs = (List.rev zs) @ [z] \<and> length zs = n" by auto
thus ?thesis by auto
qed


lemma asum_aux:
assumes lv:"n = length v" and lw:"n = length w" and le:"m \<le> n"
shows "asum (\<lambda> i. mul (v ! i) (w ! i)) m = asum (\<lambda> i. mul ((v @ [x]) ! i) ((w @ [y]) ! i)) m"
using le
proof (induct m, simp)
case (Suc k)  
 hence kn:"Suc k \<le> n" and
  ih: "k \<le> n \<Longrightarrow> asum (\<lambda>i. mul (v ! i) (w ! i)) k = asum (\<lambda>i. mul ((v @ [x]) ! i) ((w @ [y]) ! i)) k" by auto
 from kn lv have vk:"v ! k = (v @ [x]) ! k" using nth_append[where ?xs=v and ?ys="[x]"] by auto
 from kn lw have wk:"w ! k = (w @ [y]) ! k" using nth_append[where ?xs=w and ?ys="[y]"] by auto
 let ?f = "\<lambda>i. mul (v ! i) (w ! i)"
 let ?f' = "\<lambda>i. mul ((v @ [x])! i) ((w @ [y]) ! i)" 
 let ?p = "\<lambda>i s. plus (?f i) s"
 let ?p' = "\<lambda>i s. plus (?f' i) s"
(*  from lv have "\<exists> x v'. v = v' @ [x] \<and> length v' = m" using list_app_last[where?n=m] by auto
 then obtain v' x' where vis:"v = v' @ [x']" and lv':"length v' = m" by auto
 from lw have "\<exists> y w'. w = w' @ [y] \<and> length w' = m" using list_app_last[where?n=m] by auto
 then obtain w' y' where wis:"w = w' @ [y']" and lw':"length w' = m" by auto*)
 have "asum ?f (Suc k) = foldr ?p [0..<(Suc k)] zero" using asum_foldr[where ?n="Suc k"] by auto
 also have "\<dots> = foldr ?p ([0..<k] @ [k]) zero" by auto
 also have "\<dots> = foldr ?p [0..<k] (foldr ?p [k] zero)" by auto
 also have "\<dots> = foldr ?p [0..<k] (plus zero (mul (v ! k) (w ! k)))" by auto
 also have "\<dots> = plus (foldr ?p [0..<k] zero) (mul (v ! k) (w ! k))" 
  using foldr_absorb[where ?y="mul (v ! k) (w ! k)" and ?zs="[0..<k]" and ?x="zero"] by auto
 also have "\<dots> = plus (asum ?f k) (mul (v ! k) (w ! k))" using asum_foldr by auto
 also with ih kn have "\<dots> = plus (asum ?f' k) (mul (v ! k) (w ! k))" using asum_foldr by auto
 also with vk wk have "\<dots> = plus (asum ?f' k) (mul ((v @ [x]) ! k) ((w @ [y]) ! k))" using nth_append by auto
 also have "\<dots> = plus (foldr ?p' [0..<k] zero) (mul ((v @ [x]) ! k) ((w @ [y]) ! k))" using asum_foldr[where ?n=k] by auto
 also have "\<dots> = foldr ?p' [0..<k] (plus zero (mul ((v @ [x]) ! k) ((w @ [y]) ! k)))" 
  using foldr_absorb[where ?x="zero" and ?y="mul ((v @ [x]) ! k) ((w @ [y]) ! k)" and ?zs="[0..<k]"] by auto
 also have "\<dots> = foldr ?p' [0..<k] (foldr ?p' [k] zero)"  by auto
 also have "\<dots> = foldr ?p' [0..<Suc k] zero" by auto
 finally show ?case  using asum_foldr[where ?n="Suc k" and ?f = "?f'"] by auto
qed


lemma scalar_prod_asum: 
assumes wf_v:"wf_vec n v" and wf_w:"wf_vec n w"
shows "scalar_prod v w = asum (\<lambda> i. mul (v ! i) (w ! i)) n"
using assms
unfolding scalar_prod.simps
proof (induct n arbitrary: v w, simp)
(* let ?f = "\<lambda> i. mul (v ! i) (w ! i)"
let ?g = "\<lambda>xy. plus (mul (fst xy) (snd xy))"*)
case (Suc n)
 hence IH:"\<And> v w. \<lbrakk>wf_vec n v; wf_vec n w\<rbrakk> \<Longrightarrow> foldr (\<lambda>xy. plus (mul (fst xy) (snd xy))) (zip v w) zero = asum (\<lambda> i. mul (v ! i) (w ! i)) n" by auto
 from Suc have snv:"Suc n = length v" and snw:"Suc n = length w" using wf_vec.simps by auto
 hence "\<exists> x v'. v = v' @ [x] \<and> length v' = n" using list_app_last by auto
 then obtain x v' where v_is:"v = v' @ [x]" and  lv:"length v' = n" and wf_v':"wf_vec n v'" by auto
 hence vn:"v ! n = x" by auto
 from snw have "\<exists> y w'. w = w' @ [y] \<and> length w' = n" using list_app_last by auto
 then obtain y w' where w_is:"w = w' @ [y]" and wf_w': "wf_vec n w'" by auto
 hence wn:"w ! n = y" by auto
 from wf_v' wf_w' have lv:"length v' = n" and lw:"length w'=n" by auto
 let ?x_y = "Pair x y"
 let ?f = "\<lambda> i. mul (v ! i) (w ! i)"
 let ?g = "\<lambda>xy. plus (mul (fst xy) (snd xy))"
 let ?f' = "\<lambda> i. mul (v' ! i) (w' ! i)"
 let ?plus = "\<lambda> i s. plus (?f i) s"
 from lv lw have "n = length v'" and "n = length w'" and "n \<le> n" by auto
 hence "asum ?f' n = asum (\<lambda> i. mul ((v' @ [x]) ! i) ((w' @ [y]) ! i)) n" using asum_aux[where ?v="v'" and ?w="w'"] by auto
 also with v_is w_is have " \<dots> = asum ?f n" by auto
 finally have asum_eq:"asum ?f' n = asum ?f n" .
 have " asum ?f (Suc n) = foldr ?plus [0 ..<(Suc n)] zero" using asum_foldr[where ?n="Suc n"] by auto
 also have "\<dots> = foldr ?plus ([0 ..<n] @ [n]) zero" by auto
 also have "\<dots> = foldr ?plus [0 ..<n] (foldr ?plus [n] zero)" by auto
 also have "\<dots> = foldr ?plus [0 ..<n] (mul (v ! n) (w ! n))" by auto
 also have "\<dots> = plus (foldr ?plus [0 ..<n] zero) (mul (v ! n) (w ! n))" using asum_absorb0_r[where ?zs="[0..<n]"] by auto
 also with vn wn have "\<dots> = plus (asum ?f n) (mul x y)" using asum_foldr[where ?n="n"] by auto
 also with asum_eq have "\<dots> = plus (asum ?f' n) (mul x y)" by auto
 also with wf_v' wf_w' IH have "\<dots> = plus (foldr ?g (zip v' w') zero) (mul x y)" by auto
 also have "\<dots> = plus (foldr ?g (zip v' w') zero) (mul (fst ?x_y) (snd ?x_y))" by auto
 also have "\<dots> = foldr ?g (zip v' w') (plus zero (mul (fst ?x_y) (snd ?x_y)))" 
  using foldr_absorb[where ?x="zero" and ?zs="zip v' w'" and ?y="mul (fst ?x_y) (snd ?x_y)"] by auto
 also have "\<dots> = foldr ?g (zip v' w') (plus (mul (fst ?x_y) (snd ?x_y)) zero)" using plus_comm by auto
 also have "\<dots> = foldr ?g (zip v' w') (foldr ?g [?x_y] zero)" by auto
 also have "\<dots> = foldr ?g ((zip v' w') @ [?x_y]) zero" using foldr_append by auto
 also with lv lw have "\<dots> = foldr ?g (zip (v' @ [x]) (w' @ [y])) zero" using zip_append by auto
 also with v_is w_is have "\<dots> = foldr ?g (zip v w) zero" by auto
 finally show ?case by auto
qed


(* dieses Lemma wäre praktisch, um das untenstehende  mat_mul_mat_vec_assoc Lemma zu beweisen \<dots> *)
lemma scalar_prod_vec_mat_vec:
assumes wf1:"wf_vec (Suc nr) u" and wf2:"wf_mat_rc (Suc nr) (Suc nc) m" and wf3:"wf_vec (Suc nc) v"
shows "scalar_prod (rvec_mat_mul u m) v = scalar_prod u (mat_vec_mul m v)"
using assms
sorry


lemma wf_mat_vec_mul:
assumes "wf_mat_rc i j m" and "wf_vec j v"
shows "wf_vec i (mat_vec_mul m v)"
using assms by auto

lemma wf_rvec_mat_mul:
assumes "wf_vec (Suc i) v" and "wf_mat_rc (Suc i) (Suc j) m"
shows "wf_vec (Suc j) (rvec_mat_mul v m)"
using assms wf_mat_transpose by auto


lemma mat_mul_mat_vec_assoc[simp] : 
assumes wf1: "wf_mat_rc (Suc nr) (Suc nc) m1" and wf2: "wf_mat_rc (Suc nc) (Suc nr) m2" and wf3: "wf_vec (Suc nr) w"
shows "mat_vec_mul (mat_mul m1 m2) w = mat_vec_mul m1 (mat_vec_mul m2 w)"
proof-
from wf1 have wf_m1i:"\<forall> i < Suc nr. wf_vec (Suc nc) (m1 ! i)" and "length m1 = (Suc nr)" by auto
hence "\<forall> i < length m1. wf_vec (Suc nc) (m1 ! i)" by auto
hence wfm1rows:"\<forall> v \<in> set m1. wf_vec (Suc nc) v" using all_set_conv_all_nth[where ?P="\<lambda>x. wf_vec (Suc nc) x"] by auto
from wf_m1i wf2 have wf_m1m2:"\<forall> i < Suc nr. wf_vec (Suc nr) (rvec_mat_mul (m1 ! i) m2)" using wf_rvec_mat_mul by auto
from wf2 wf3 have wfm2w:"wf_vec (Suc nc)  (mat_vec_mul m2 w)" using wf_mat_vec_mul by auto
have "mat_vec_mul (mat_mul m1 m2) w = mat_vec_mul (map (\<lambda> v. rvec_mat_mul v m2) m1) w" by auto
also have "\<dots> = map (scalar_prod w) (map (\<lambda> v. rvec_mat_mul v m2) m1)" by auto
also have "\<dots> = map (\<lambda> v. scalar_prod w (rvec_mat_mul v m2)) m1" using map_compose[symmetric] by auto
also with wf_m1m2 wf3 have "\<dots> = map (\<lambda> v. scalar_prod (rvec_mat_mul v m2) w) m1" using scalar_prod_comm by auto
also with wfm1rows wf2 wf3 have "\<dots> = map (\<lambda> v. scalar_prod v (mat_vec_mul m2 w)) m1" using scalar_prod_vec_mat_vec[where ?nr="nc" and ?nc="nr" and ?m="m2" and ?v="w"] by auto
also with wfm1rows wfm2w have "\<dots> = map (scalar_prod (mat_vec_mul m2 w)) m1" using scalar_prod_comm by auto
finally show ?thesis by auto
qed
*)
end

end