(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory RedPairImpl imports PolyOrder
begin

datatype arctic = MinInfty | Nat_arc nat

datatype 'f redpair = Nat_carrier "('f,nat)lpoly_interL" | Int_carrier "('f,int)lpoly_interL" | Arctic_carrier "('f,arctic)lpoly_interL"

fun plus_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> arctic"
where "plus_arctic MinInfty y = y"
   |  "plus_arctic x MinInfty = x"
   |  "plus_arctic (Nat_arc x) (Nat_arc y) = (Nat_arc (max x y))"

fun mult_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> arctic"
where "mult_arctic MinInfty y = MinInfty"
   |  "mult_arctic x MinInfty = MinInfty"
   |  "mult_arctic (Nat_arc x) (Nat_arc y) = (Nat_arc (x + y))"

fun geq_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> bool"
where "geq_arctic x MinInfty = True"
   |  "geq_arctic MinInfty (Nat_arc _) = False"
   |  "geq_arctic (Nat_arc x) (Nat_arc y) = (x \<ge> y)"

fun gt_arctic :: "arctic \<Rightarrow> arctic \<Rightarrow> bool"
where "gt_arctic MinInfty _ = False"
   |  "gt_arctic (Nat_arc _) MinInfty = True"
   |  "gt_arctic (Nat_arc x) (Nat_arc y) = (x > y)"

instantiation arctic :: sshow
begin
fun sshow_arctic :: "arctic \<Rightarrow> sstring"
where "sshow_arctic (Nat_arc x) = sshow x"
   |  "sshow_arctic MinInfty = ''-inf''"
instance ..
end

definition nat_name :: "sstring" where "nat_name = ''naturals''"
definition int_name :: "sstring" where "int_name = ''naturals with negative constants''"
definition arctic_name :: "sstring" where "arctic_name = ''arctic semiring''"

interpretation nat_poly_carrier: poly_carrier "\<lambda> x. True" "0 :: nat" "op +" "op *" "op \<ge>" "op >" True
  by (unfold_locales, auto simp: nat_distrib)

interpretation int_poly_carrier: poly_carrier "\<lambda> x. True" "0 :: int" "op +" "op *" "op \<ge>" "op >" True
  by (unfold_locales, auto simp: int_distrib Ring_and_Field.mult_right_mono Ring_and_Field.mult_left_mono)

interpretation arctic_poly_carrier: poly_carrier "\<lambda> x. True" MinInfty plus_arctic mult_arctic geq_arctic gt_arctic False
proof (unfold_locales, auto)
  fix x y
  show "plus_arctic x y = plus_arctic y x"
    by (cases x, cases y, auto, cases y, auto)
next
  fix x y z
  show "plus_arctic x (plus_arctic y z) = plus_arctic (plus_arctic x y) z"
    by (cases x, auto, cases y, auto, cases z, auto)
next
  fix x y z
  show "mult_arctic x (mult_arctic y z) = mult_arctic (mult_arctic x y) z"
    by (cases x, auto, cases y, auto, cases z, auto)
next
  fix x 
  show "mult_arctic x MinInfty = MinInfty"
    by (cases x, auto)
next
  fix x y z
  show "mult_arctic x (plus_arctic y z) = plus_arctic (mult_arctic x y) (mult_arctic x z)"
    by (cases x, auto, cases y, auto, cases z, auto)
next
  fix x y z
  show "mult_arctic (plus_arctic x y) z = plus_arctic (mult_arctic x z) (mult_arctic y z)"
    by (cases x, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z
  assume "geq_arctic x y" and "gt_arctic y z"
  thus "gt_arctic x z"
    by (cases x, cases y, auto, cases y, auto, cases z, auto)
next 
  fix x y z
  assume "gt_arctic x y" and "geq_arctic y z"
  thus "gt_arctic x z"
    by (cases x, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z
  assume "geq_arctic x y" 
  thus "geq_arctic (plus_arctic x z) (plus_arctic y z)"
    by (cases x, cases y, cases z, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z
  assume "geq_arctic x y" 
  thus "geq_arctic (mult_arctic x z) (mult_arctic y z)"
    by (cases x, cases y, cases z, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z
  assume "geq_arctic y z" 
  thus "geq_arctic (mult_arctic x y) (mult_arctic x z)"
    by (cases x, cases y, cases z, auto, cases y, cases z, auto, cases z, auto)
next
  fix x
  show "geq_arctic x x" 
    by (cases x, auto)
next
  fix x y z
  assume "geq_arctic x y" and "geq_arctic y z"
  thus "geq_arctic x z"
    by (cases x, cases y, auto, cases y, cases z, auto, cases z, auto)
next
  fix x y z u
  assume "gt_arctic x y" and "gt_arctic z u"
  thus "gt_arctic (plus_arctic x z) (plus_arctic y u)"
    by (cases x, simp, cases z, simp, cases y, cases u, auto, cases u, auto)
qed


interpretation nat_poly_order_carrier : poly_order_carrier "\<lambda> x. True" "0 :: nat" "op +" "op *" "op \<ge>" "op >" True 1 id nat_name
proof (unfold_locales, auto)
  have "{(x,y). (x :: nat) < y} = ({(x,y). (x :: nat) > y}^-1)" by auto
  with wf_less have wf: "wf ({(x,y). (x :: nat) > y}^-1)" by simp 
  show "SN {(x :: nat,y). x > y}" 
    by (rule wf_imp_SN, rule wf)
qed

interpretation int_poly_order_carrier : poly_order_carrier "\<lambda> x. True" "0 :: int" "op +" "op *" "op \<ge>" "op >" True 1 "max 0" int_name
proof (unfold_locales, auto)
  have "{(x,y). 0 \<le> x \<and> (x :: int) < y} = ({(x,y). y \<ge> 0 \<and> (x :: int) > y}^-1)" by auto
  with wf_int_ge_less_than[where d = 0] have wf: "wf ({(x,y). y \<ge> 0 \<and> (x :: int) > y}^-1)" unfolding int_ge_less_than_def 
    by simp
  show "SN {(x :: int,y). 0 \<le> y  \<and> x > y}"
    by (rule wf_imp_SN, rule wf)
qed

fun get_arctic_nat :: "arctic \<Rightarrow> nat"
where "get_arctic_nat (Nat_arc n) = n"

interpretation arctic_poly_order_carrier : poly_order_carrier "\<lambda> x. True" MinInfty plus_arctic mult_arctic geq_arctic gt_arctic False "Nat_arc 0" id arctic_name
proof (unfold_locales, auto)
  fix x 
  show "mult_arctic (Nat_arc 0) x = x" by (cases x, auto)
next
  fix x 
  show "mult_arctic x (Nat_arc 0) = x" by (cases x, auto)
next
  show "SN {(x,y). gt_arctic x y}"
    unfolding SN_ars_def SN_elt_def
  proof (rule allI, rule notI)
    fix x
    assume "\<exists> f . f 0 = x \<and> (\<forall> i. (f i, f (Suc i)) \<in> {(x,y). gt_arctic x y})"
    from this obtain f where "f 0 = x" and seq: "\<forall> i. (f i, f (Suc i)) \<in> {(x,y). gt_arctic x y}" by auto
    from seq have steps: "\<forall> i. gt_arctic (f i) (f (Suc i))" by auto
    let ?g = "\<lambda> i. get_arctic_nat (f i)"
    have "\<forall> i. ?g i  > ?g (Suc i)"
    proof
      fix i
      from steps have i: "gt_arctic (f i) (f (Suc i))" and si: "gt_arctic (f (Suc i)) (f (Suc (Suc i)))" by auto
      from i obtain n where fi: "f i = Nat_arc n" by (cases "f i", auto)
      from si obtain m where fsi: "f (Suc i) = Nat_arc m" by (cases "f (Suc i)", auto)
      from i fi fsi have "n > m" by auto
      with fi fsi
      show "?g i > ?g (Suc i)" by auto
    qed
    from this obtain g where "\<forall> i. (op > :: nat \<Rightarrow> nat \<Rightarrow> bool) (g i) (g (Suc i))" by auto
    hence contra: "\<exists> f. f 0 = g 0 \<and> (\<forall> i. (f i, f (Suc i)) \<in> {(x,y). x > y})" by auto
    from nat_poly_order_carrier.SN
    have SN: "SN {(x,y :: nat). x > y}" by auto
    with contra show False unfolding SN_ars_def SN_elt_def by blast 
  qed
qed

interpretation nat_mono_poly_order_carrier : mono_poly_order_carrier "\<lambda> x. True" "0 :: nat" "op +" "op *" "op \<ge>" "op >" True 1 id nat_name "\<lambda> x. x \<noteq> 0"
  by (unfold_locales, auto)

interpretation int_mono_poly_order_carrier : mono_poly_order_carrier "\<lambda> x. True" "0 :: int" "op +" "op *" "op \<ge>" "op >" True 1 "max 0" int_name "\<lambda> x. x > 0"
  by (unfold_locales, auto simp: Ring_and_Field.mult_strict_left_mono)

abbreviation nat_components :: "nat poly_components"
where "nat_components \<equiv> ((\<lambda> x. True) , 0 , 1 , op +, op *, id, op \<ge>, op >, True, nat_name)"

abbreviation int_components :: "int poly_components"
where "int_components \<equiv> ((\<lambda> x. True) , 0 , 1 , op +, op *, max 0, op \<ge>, op >, True, int_name)"

abbreviation arctic_components :: "arctic poly_components"
where "arctic_components \<equiv> ((\<lambda> x. True), MinInfty, Nat_arc 0, plus_arctic, mult_arctic, id, geq_arctic, gt_arctic, False, arctic_name)"

fun create_rule_removal_proc :: "('f :: sshow)redpair \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow>
        sstring list monad"
where "create_rule_removal_proc (Int_carrier I) = create_rule_removal_proc_comp int_components (\<lambda> x. x > 0) I"
  |   "create_rule_removal_proc (Nat_carrier I) = create_rule_removal_proc_comp nat_components (\<lambda> x. x \<noteq> 0) I"
  |   "create_rule_removal_proc (Arctic_carrier I) = (\<lambda> _ _. error [''arctic not monotonic for TRSs''])"

fun create_redpair_proc :: "('f :: sshow)redpair \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_redpair_proc (Int_carrier I) = create_redpair_proc_comp int_components I"
  |   "create_redpair_proc (Nat_carrier I) = create_redpair_proc_comp nat_components I"
  |   "create_redpair_proc (Arctic_carrier I) = create_redpair_proc_comp arctic_components I"


fun create_mono_redpair_proc :: "('f :: sshow)redpair \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_mono_redpair_proc (Int_carrier I) = create_mono_redpair_proc_comp int_components (\<lambda> x. x > 0) I"
  |   "create_mono_redpair_proc (Nat_carrier I) = create_mono_redpair_proc_comp nat_components (\<lambda> x. x \<noteq> 0) I"
  |   "create_mono_redpair_proc (Arctic_carrier I) = (\<lambda> _ _ _ _. error [''arctic not monotonic for TRSs''])"


fun create_redpair_ur_proc :: "('f :: sshow)redpair \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_redpair_ur_proc (Int_carrier I) = create_redpair_ur_proc_comp int_components I"
  |   "create_redpair_ur_proc (Nat_carrier I) = create_redpair_ur_proc_comp nat_components I"
  |   "create_redpair_ur_proc (Arctic_carrier I) = create_redpair_ur_proc_comp arctic_components I"

fun create_mono_redpair_ur_proc :: "('f :: sshow)redpair \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_mono_redpair_ur_proc (Int_carrier I) = create_mono_redpair_ur_proc_comp int_components (\<lambda> x. x > 0) I"
  |   "create_mono_redpair_ur_proc (Nat_carrier I) = create_mono_redpair_ur_proc_comp nat_components (\<lambda> x. x \<noteq> 0) I"
  |   "create_mono_redpair_ur_proc (Arctic_carrier I) = (\<lambda> _ _ _ _ _ _. error [''arctic not monotonic for TRSs''])"

fun create_redpair_ur_af_proc :: "('f :: sshow)redpair \<Rightarrow> ('f,'v :: sshow)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
  ('f \<Rightarrow> ('f,'v)trsL)option \<Rightarrow> ('f,'v)trsL \<Rightarrow> 
        sstring list monad"
where "create_redpair_ur_af_proc (Int_carrier I) = create_redpair_ur_af_proc_comp int_components I"
  |   "create_redpair_ur_af_proc (Nat_carrier I) = create_redpair_ur_af_proc_comp nat_components I"
  |   "create_redpair_ur_af_proc (Arctic_carrier I) = create_redpair_ur_af_proc_comp arctic_components I"


lemma create_rule_removal_proc: 
  assumes check: "isOK(create_rule_removal_proc I R rR)"
  and SN: "SN (rstep (set rR))"
  shows "SN (rstep (set R))"
proof (cases I)
  case (Nat_carrier nI)
  hence check1: "isOK(create_rule_removal_proc_comp nat_components (\<lambda> x. x \<noteq> 0) nI R rR)" using check by auto
  show ?thesis by (rule nat_mono_poly_order_carrier.create_rule_removal_proc_comp_sound, rule check1, rule SN)
next
  case (Int_carrier nI)
  hence check1: "isOK(create_rule_removal_proc_comp int_components (\<lambda> x. x > 0) nI R rR)" using check by auto
  show ?thesis by (rule int_mono_poly_order_carrier.create_rule_removal_proc_comp_sound, rule check1, rule SN)
next
  case (Arctic_carrier nI) thus ?thesis using check by auto
qed


lemma create_redpair_proc: "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp nat_components nI P rP R))" by auto
  show ?thesis by (simp only: id,  rule nat_poly_order_carrier.redpair_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp int_components nI P rP R))" by auto
  show ?thesis by (simp only: id,  rule int_poly_order_carrier.redpair_proc_comp_ci)
next
  case (Arctic_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_proc I P rP R)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_proc_comp arctic_components nI P rP R))" by auto
  show ?thesis by (simp only: id,  rule arctic_poly_order_carrier.redpair_proc_comp_ci)
qed


lemma create_mono_redpair_proc: "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp nat_components (\<lambda> x. x \<noteq> 0) nI P rP R rR))" by auto
  show ?thesis by (simp only: id, rule nat_mono_poly_order_carrier.mono_redpair_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc I P rP R rR)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_proc_comp int_components (\<lambda> x. x > 0) nI P rP R rR))" by auto
  show ?thesis by (simp only: id, rule int_mono_poly_order_carrier.mono_redpair_proc_comp_ci)
next
  case (Arctic_carrier nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
qed

lemma create_redpair_ur_proc: "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp nat_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule nat_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp int_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule int_poly_order_carrier.redpair_ur_proc_comp_ci)
next
  case (Arctic_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_proc_comp arctic_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule arctic_poly_order_carrier.redpair_ur_proc_comp_ci)
qed


lemma create_redpair_ur_af_proc: "chain_identifying_onlyDP_proc_impl (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp nat_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule nat_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next 
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp int_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule int_poly_order_carrier.redpair_ur_af_proc_comp_ci)
next
  case (Arctic_carrier nI)
  hence id: "(\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc I P rP R rm ur)) = 
    (\<lambda> P R rm rrm rP. isOK(create_redpair_ur_af_proc_comp arctic_components nI P rP R rm ur))" by auto
  show ?thesis by (simp only: id,  rule arctic_poly_order_carrier.redpair_ur_af_proc_comp_ci)
qed

lemma create_mono_redpair_ur_proc: "chain_identifying_proc_impl (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur))"
proof (cases I)
  case (Nat_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp nat_components (\<lambda> x. x \<noteq> 0) nI P rP R rm rR ur))" by auto
  show ?thesis by (simp only: id, rule nat_mono_poly_order_carrier.mono_redpair_ur_proc_comp_ci)
next
  case (Int_carrier nI)
  hence id: "(\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc I P rP R rm rR ur)) = 
    (\<lambda> P R rm rP rR. isOK(create_mono_redpair_ur_proc_comp int_components (\<lambda> x. x > 0) nI P rP R rm rR ur))" by auto
  show ?thesis by (simp only: id, rule int_mono_poly_order_carrier.mono_redpair_ur_proc_comp_ci)
next
  case (Arctic_carrier nI) thus ?thesis unfolding chain_identifying_proc_impl_def by auto
qed

  
end